# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Mean square displacement."""

import logging
import numpy

from .helpers import linear_grid
from .correlation import Correlation, gcf_offset
from .helpers import setup_t_grid

__all__ = ['MeanSquareDisplacement']

_log = logging.getLogger(__name__)


class MeanSquareDisplacement(Correlation):
    """
    Mean square displacement.

    If the time grid `tgrid` is None, the latter is redefined in a way
    controlled by the variable `rmax`. If `rmax` is negative
    (default), the time grid is linear between 0 and half of the
    largest time in the trajectory. If `rmax` is positive, the time
    grid comprises `tsamples` entries linearly spaced between 0 and
    the time at which the square root of the mean squared displacement
    reaches `rmax`.

    Additional parameters:
    ----------------------

    - sigma: value of the interparticle distance (usually 1.0). It is
    used to limit the fit range to extract the diffusion coefficient
    and to determine the diffusion time
    """

    symbol = 'msd'
    short_name = 'dr^2(t)'
    long_name = 'mean square displacement'
    phasespace = 'pos-unf'

    def __init__(self, trajectory, tgrid=None, rmax=-1.0, norigins=None,
                 tsamples=30, sigma=1.0, fix_cm=False, no_offset=False):
        self.rmax = rmax
        self.sigma = sigma
        self.tsamples = tsamples
        self._norigins = norigins
        # Offsets in time grid are only relevant with periodic blocks
        # Use the no_offset parameter to disable them
        self._no_offset = no_offset
        Correlation.__init__(self, trajectory, tgrid, norigins=norigins, fix_cm=fix_cm)        
        
    def _compute(self):
        # We postpone time grid definition to compute to avoid
        # unfolding the trajectory twice when targeting the rmsd

        def msd(x, y):
            return numpy.sum((x-y)**2) / float(x.shape[0])

        if self.grid is None:
            t_max = self.trajectory.total_time
            if self.rmax > 0.0:
                msd_total = msd(self._pos_unf[-1], self._pos_unf[0])
                frac = self.rmax**2 / msd_total
                t_target = frac * t_max
                self.grid = linear_grid(0.0, min(t_max, t_target), self.tsamples)
            else:
                self.grid = linear_grid(0.0, t_max, self.tsamples)

        # Setup time grid
        # We disable offsets if only one time origin is requested or if no_offset is True
        self._discrete_tgrid = setup_t_grid(self.trajectory, self.grid,
                                            offset=self._norigins != '1' and not self._no_offset)
        # Note that the grid is redefined
        self.grid, self.value = gcf_offset(msd, self._discrete_tgrid, self.skip,
                                           self.trajectory.steps, self._pos_unf)
        # Update grid to real time
        self.grid = [ti * self.trajectory.timestep for ti in self.grid]

    def analyze(self):
        # Get the time when MSD equals sigma**2
        try:
            from .helpers import feqc
            self.analysis['diffusive time tau_D'] = feqc(self.grid, self.value, self.sigma**2)[0]
        except ValueError:
            self.analysis['diffusive time tau_D'] = None

        where = numpy.array(self.value) > self.sigma**2
        if sum(where) < 2:
            _log.warn('could not fit MSD: not enough data above sigma')
            return

        from .helpers import linear_fit
        diffusion = linear_fit(numpy.array(self.grid)[where],
                               numpy.array(self.value)[where])
        ndim = self.trajectory.read(0).number_of_dimensions
        self.analysis['diffusion coefficient D'] = diffusion[0] / (2*ndim)



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())

from .correlation import *
from .partial import *
from .filter import Filter
from .susceptibility import Susceptibility

# Real space correlation functions
from .alpha2 import *
from .chi4t import *
from .gr import *
from .msd import *
from .qt import *
from .vacf import *
from .ba import *

# Real space correlation functions
from .fkt import *
from .ik import *
from .s4kt import *
from .sk import *





# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""
Decorator to filter correlation functions by arbitrary condition
on particle properties.

It uses filters internally.

For 2-body correlations we expect up to two logical conditions. If
there is only one, we filter only the first group of particle, the
second one includes all the particles of the system. Otherwise we
filter two subsets of particles accordingly.
"""

import logging

from .helpers import filter_species

__all__ = ['Filter']
_log = logging.getLogger(__name__)


def Filter(correlation, condition):

    def _filter_generic(system, condition):
        import copy
        s = copy.copy(system)
        s.particle = []
        for particle in system.particle:
            if eval(condition, particle.__dict__):
                s.particle.append(particle)
        return s

    tag = condition.replace(' ', '')
    tag = tag.replace('and', '_')
    tag = tag.replace('or', '+')
    tag = tag.replace('"', '')
    tag = tag.replace("'", '')
    tag = tag.replace('==', '')
    tag = tag.replace('!=', '.not')
    tag = tag.replace(',', '-')
    correlation.tag = tag
    correlation.tag_description = condition

    if correlation.nbodies == 1:
        correlation.add_filter(_filter_generic, condition)

    elif correlation.nbodies == 2:
        # We expect up to two logical conditions
        conditions = condition.split(',')

        # If there is only one, we filter only the first group of
        # particles, the second one includes all the particles of the
        # system, hence the identity function
        if len(conditions) == 1:
            correlation.add_filter(_filter_generic, conditions[0])
            correlation.add_filter(lambda x: x)
        elif len(conditions) == 2:
            # If there are two, we filter two subsets of particles
            # accordingly
            correlation.add_filter(_filter_generic, conditions[0])
            correlation.add_filter(_filter_generic, conditions[1])
        else:
            raise ValueError('too many conditions for a 2 body-correlation function')

    return correlation



import os
import argparse

from ._version import __version__ as __bare_version

# Version
try:
    from ._commit import __commit__, __date__
    __version__ = '%s+%s (%s)' % (__bare_version, __commit__, __date__)
except ImportError:
    __commit__ = ""
    __date__ = ""
    __version__ = __bare_version

# Global variables
pp_output_path = '{trajectory.filename}.pp.{symbol}.{tag}'
pp_trajectory_format = None

# Help formatter
os.environ['COLUMNS'] = "100"

class CustomHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def __init__(self, prog, *args, **kwargs):
        argparse.RawDescriptionHelpFormatter.__init__(self, prog,
                                                      indent_increment=2,
                                                      max_help_position=60,
                                                      width=None)

    def _get_help_string(self, action):
        help = action.help
        if '%(default)' not in action.help:
            if action.default is not argparse.SUPPRESS:
                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    help += ' [default: %(default)s]'
        return help

    def __add_whitespace(self, idx, iWSpace, text):
        if idx == 0:
            return text
        return (" " * iWSpace) + text

    def _split_lines(self, text, width):
        import re
        import textwrap as _textwrap
        textRows = text.splitlines()
        for idx,line in enumerate(textRows):
            search = re.search('\s*[0-9\-]{0,}\.?\s*', line)
            if line.strip() is "":
                textRows[idx] = " "
            elif search:
                lWSpace = search.end()
                lines = [self.__add_whitespace(i,lWSpace,x) for i,x in enumerate(_textwrap.wrap(line, width))]
                textRows[idx] = lines

        return [item for sublist in textRows for item in sublist]





# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Four-point dynamic susceptibility."""

import logging
import numpy

from .helpers import logx_grid
from .correlation import Correlation
from .helpers import setup_t_grid, ifabsmm
from .qt import self_overlap
from .progress import progress

__all__ = ['Chi4SelfOverlap', 'Chi4SelfOverlapOptimized']

_log = logging.getLogger(__name__)


class Chi4SelfOverlap(Correlation):
    """
    Four-point dynamic susceptibility from the time-dependent self
    overlap function.

    Parameters:
    -----------

    - a: distance parameter entering the Heaviside function in the
    overlap calculation
    """

    symbol = 'chi4qs'
    short_name = 'chi_4(t)'
    long_name = 'dynamic susceptibility of self overlap'
    phasespace = 'pos-unf'

    def __init__(self, trajectory, tgrid=None, norigins=-1, a=0.3,
                 tsamples=60):
        Correlation.__init__(self, trajectory, tgrid, norigins=norigins)
        if tgrid is None:
            self.grid = logx_grid(0.0, self.trajectory.total_time * 0.75, tsamples)
        self._discrete_tgrid = setup_t_grid(self.trajectory, self.grid, offset=norigins != '1')
        self.a_square = a**2
        self.average = Correlation(self.trajectory, self.grid)
        self.average.short_name = 'Q^u(t)'
        self.average.symbol = 'qsu'
        self.average.long_name = 'Average of self overlap, not normalized'
        self.variance = Correlation(self.trajectory, self.grid)
        self.variance.short_name = 'Q_2^u(t)'
        self.variance.symbol = 'q2su'
        self.variance.long_name = 'Variance of self overlap, not normalized'

    def _compute(self):
        def f(x, y):
            return self_overlap(x, y, side, self.a_square).sum()
        
        side = self.trajectory.read(0).cell.side
        self.grid = []
        # At this stage, we can copy the tags
        self.average.tag, self.variance.tag = self.tag, self.tag
        for off, i in progress(self._discrete_tgrid):
            A, A2, cnt = 0.0, 0.0, 0
            for i0 in range(off, len(self._pos_unf)-i-self.skip, self.skip):
                w = f(self._pos_unf[i0], self._pos_unf[i0 + i])
                A2 += w**2
                A += w
                cnt += 1
            dt = self.trajectory.steps[off+i] - self.trajectory.steps[off]
            if cnt > 0:
                A_av, A2_av = A / cnt, A2 / cnt
            else:
                A_av, A2_av = 0, 0
            self.grid.append(dt * self.trajectory.timestep)
            self.value.append((A2_av - A_av**2) / self._pos_unf[0].shape[0])
            self.average.value.append(A_av)
            self.variance.value.append(A2_av)
        self.average.grid, self.variance.grid = self.grid, self.grid

    def write(self):
        # We subclass this to also write down qsu and qsu2
        super(Chi4SelfOverlap, self).write()
        self.average.write()
        self.variance.write()

    def analyze(self):
        try:
            self.analysis['peak time tau_star'], self.analysis['peak height chi4_star'] = ifabsmm(self.grid, self.value)[1]
        except ZeroDivisionError:
            _log.warning('could not find maximum')


class Chi4SelfOverlapOptimized(Chi4SelfOverlap):
    """
    Four-point dynamic susceptibility from the time-dependent self
    overlap function.

    Optimized version using fortran 90 extension.
    """

    def _compute(self):
        try:
            from atooms.postprocessing.realspace_wrap import realspace_module
        except ImportError:
            _log.error('f90 wrapper missing or not functioning')
            raise

        def f(x, y):
            return realspace_module.self_overlap(x, y, numpy.array(self.a_square))

        self.grid = []
        self.average.tag, self.variance.tag = self.tag, self.tag
        for off, i in progress(self._discrete_tgrid):
            A, A2, cnt = 0.0, 0.0, 0
            for i0 in range(off, len(self._pos_unf)-i-self.skip, self.skip):
                w = f(self._pos_unf[i0], self._pos_unf[i0+i])
                A2 += w**2
                A += w
                cnt += 1
            dt = self.trajectory.steps[off+i] - self.trajectory.steps[off]
            A_av, A2_av = A/cnt, A2/cnt
            self.grid.append(dt * self.trajectory.timestep)
            self.value.append((A2_av - A_av**2) / self._pos_unf[0].shape[0])
            self.average.value.append(A_av)
            self.variance.value.append(A2_av)
        self.average.grid, self.variance.grid = self.grid, self.grid



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Bond angle distribution."""

import math
import logging

import numpy

from .helpers import linear_grid
from .correlation import Correlation
from .progress import progress

__all__ = ['BondAngleDistribution']

_log = logging.getLogger(__name__)


def _default_rcut(th):
    """
    Look for the first minimum in the partial g(r)
    """
    from atooms.system.particle import distinct_species
    from atooms.postprocessing.partial import Partial
    from atooms.postprocessing import RadialDistributionFunction
    from .helpers import ifabsmm

    ids = distinct_species(th[0].particle)
    gr = Partial(RadialDistributionFunction, ids, th, dr=0.1)
    gr.do(update=False)
    rcut = {}
    for isp in ids:
        for jsp in ids:
            # First find absolute maximum
            _, m = ifabsmm(list(gr.partial[(isp, jsp)].grid),
                           list(gr.partial[(isp, jsp)].value))
            # Then look for first minimum after the maximum
            for i in range(len(gr.partial[(isp, jsp)].grid)):
                if gr.partial[(isp, jsp)].grid[i] >= m[0]:
                    delta = gr.partial[(isp, jsp)].value[i+1] - gr.partial[(isp, jsp)].value[i]
                    if delta >= 0:
                        rcut[(isp, jsp)] = gr.partial[(isp, jsp)].grid[i]
                        break
    
    return rcut

class BondAngleDistribution(Correlation):
    """
    Bond angle distribution function.
    """

    nbodies = 2
    symbol = 'ba'
    short_name = 'D(theta)'
    long_name = 'bond angle distribution'
    phasespace = ['pos', 'ids']

    def __init__(self, trajectory, norigins=None, dtheta=4.0, rcut=None):
        Correlation.__init__(self, trajectory, None, norigins=norigins)
        self.grid = linear_grid(0.0, 180.0, dtheta)  # reassign grid anyway
        self.rcut = rcut

    def _compute(self):
        from atooms.trajectory.decorators import change_species
        from atooms.postprocessing.realspace_wrap import compute
        from atooms.system.particle import distinct_species

        hist_all = []
        hist_one = numpy.ndarray(len(self.grid), dtype=numpy.int32)
        origins = range(0, len(self.trajectory), self.skip)
        dtheta = self.grid[1] - self.grid[0]

        # Setup array of cutoff distances based on g(r) calculated for
        # the whole trajectory.
        # We store the cutoffs into a (nsp, nsp) array 
        # where the index follows the alphabetic order of species names
        ids = distinct_species(self.trajectory[0].particle)
        if self.rcut is None:
            rcut = _default_rcut(self.trajectory)
            self.rcut = numpy.ndarray((len(ids), len(ids)))
            for species_pair in rcut:
                self.rcut[ids.index(species_pair[0]), ids.index(species_pair[1])] = rcut[species_pair]
        else:
            self.rcut = numpy.array(self.rcut)

        for isp in range(len(ids)):
            for jsp in range(len(ids)):
                self.analysis['cutoff distance {}-{}'.format(isp, jsp)] = self.rcut[isp, jsp]

        for i in progress(origins):
            system = self.trajectory[i]
            side = system.cell.side  # this should not be affected by filters
            #system = change_species(system, 'F')  # species are in fortran style
            #ids = numpy.array(system.dump('spe'), dtype=numpy.int32)
            nn = numpy.array(0, dtype=numpy.int32)
            neighbors = numpy.ndarray(50, dtype=numpy.int32)
            for idx in range(len(self._pos_0[i])):
                isp = self._ids_0[i][idx]
                compute.neighbors('C', side, self._pos_0[i][idx],
                                  self._pos_1[i].transpose(),
                                  self._ids_1[i],
                                  self.rcut[isp, :],
                                  nn, neighbors)
                #print idx, self._pos_1[i].shape, neighbors[0: nn], set(self._ids_1[i])
                compute.bond_angle(self._pos_0[i][idx, :],
                                   self._pos_1[i].transpose(),
                                   neighbors[0: nn], side, dtheta,
                                   hist_one)
                hist_all.append(hist_one.copy())

        # Normalization
        hist = numpy.sum(hist_all, axis=0)
        norm = float(numpy.sum(hist[:-1]))
        self.grid = (numpy.array(self.grid[:-1]) + numpy.array(self.grid[1:])) / 2.0
        self.value = hist[:-1] / (norm * dtheta)



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Four-point dynamic structure factor."""

import numpy

from .fourierspace import FourierSpaceCorrelation, expo_sphere
from .helpers import setup_t_grid
from .qt import self_overlap
from .progress import progress

__all__ = ['S4ktOverlap']


class S4ktOverlap(FourierSpaceCorrelation):
    """
    Four-point dynamic structure factor from time-dependent self overlap.

    See the documentation of the `FourierSpaceCorrelation` base class
    for information on the instance variables.
    """

    symbol = 's4kt'
    short_name = 'S_4(t,k)'
    long_name = '4-point dynamic structure factor from self overlap'
    phasespace = ['pos', 'pos-unf']

    # TODO: refactor a S4k base correlation that forces to implement tabulat method (e.g. overlap, Q_6, voronoi ...)
    # TODO: should we drop this instead and rely on F(k,t) with grandcanonical

    def __init__(self, trajectory, tgrid, kgrid=None, norigins=-1,
                 nk=20, dk=0.1, a=0.3, kmin=1.0, kmax=10.0, ksamples=10):
        FourierSpaceCorrelation.__init__(self, trajectory, [tgrid, kgrid], norigins,
                                         nk, dk, kmin, kmax, ksamples)
        # Setup time grid
        self._discrete_tgrid = setup_t_grid(self.trajectory, tgrid, offset=norigins != '1')
        self.a_square = a**2

    def _tabulate_W(self, kgrid, selection, t_off, t, skip):
        """
        Tabulate W
        """
        side = self.trajectory[0].cell.side
        kmax = max(self.kvector.keys()) + self.dk
        nt = range(t_off, len(self._pos)-t, skip)
        W = {}
        for i_0, t_0 in enumerate(nt):
            expo = expo_sphere(self.k0, kmax, self._pos[t_0])
            for kk, knorm in enumerate(kgrid):
                for i in selection[kk]:
                    ik = self.kvector[knorm][i]
                    if ik not in W:
                        W[ik] = numpy.ndarray(len(nt), dtype=complex)
                    W[ik][i_0] = numpy.sum(self_overlap(self._pos_unf[t_0], self._pos_unf[t_0+t], side, self.a_square) *
                                          expo[..., 0, ik[0]] * expo[..., 1, ik[1]] * expo[..., 2, ik[2]])
        return W

    def _compute(self):
        # We expect there is only one time in tgrid.
        # We could easily workaround it by outer looping over i
        # We do not expect to do it for many times (typically we show S_4(k,tau_alpha) vs k)
        dt = []
        self.value = []
        for off, i in progress(self._discrete_tgrid):
            # As for fkt
            W = self._tabulate_W(self.kgrid, self.selection, off, i, self.skip)

            # Compute vriance of W
            w_av = [complex(0., 0.) for _ in self.kgrid]
            w2_av = [complex(0., 0.) for _ in self.kgrid]
            for kk, knorm in enumerate(self.kgrid):
                for j in self.selection[kk]:
                    ik = self.kvector[knorm][j]
                    # Comupte |<W>|^2  and <W W^*>
                    w_av[kk] = numpy.average(W[ik])
                    w2_av[kk] = numpy.average(W[ik] * W[ik].conjugate())

            # Normalization
            npart = self._pos[0].shape[0]
            dt.append(self.trajectory.timestep * (self.trajectory.steps[off+i] - self.trajectory.steps[off]))
            self.value.append([float(w2_av[kk] - (w_av[kk]*w_av[kk].conjugate())) / npart for kk in range(len(self.grid[1]))])
        self.grid[1] = self.kgrid
        self.grid[0] = dt



"""
Progress bar support.

We support progress bars via tqdm. If the module is not found, we
provide a fallback class that monkey-patches it.
"""

import sys

# See https://stackoverflow.com/questions/3160699/python-progress-bar for simple fallbacks

# These module level variables can be tweaked at run time
active = False
ncols = 80
bar_format = '# {l_bar} {bar} | {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_noinv_fmt}{postfix}]'


class NoProgressBar(object):

    """Fallback progress bar that monkey patches tqdm"""

    def __init__(self, iterable=None, *args, **kwargs):
        self.iterable = iterable

    def update(self, value):
        pass

    def close(self):
        pass

    def __len__(self):
        return len(self.iterable)

    def __iter__(self):
        iterable = self.iterable
        for obj in iterable:
            yield obj

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()
        return False

try:

    from tqdm import tqdm

    class CustomProgressBar(tqdm):

        """Slightly customized tqdm progress bar"""

        def __init__(self, *args, **kwargs):
            tqdm.__init__(self, disable=not active,
                          bar_format=bar_format, ncols=ncols, file=sys.stdout, *args, **kwargs)

        def update(self, n=1):
            if not self.disable:
                tqdm.update(self, n-self.n)

    progress = CustomProgressBar

except ImportError:
    progress = NoProgressBar



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""
Fake decorator to compute partial correlation functions.

It uses filters internally.
"""

import logging

from .helpers import filter_species

_log = logging.getLogger(__name__)


class Partial(object):

    def __init__(self, corr_cls, species, *args, **kwargs):
        """The first positional argument must be the trajectory instance."""
        # Instantiate correlation objects
        # with args passed upon construction
        self.partial = {}
        self.nbodies = corr_cls.nbodies
        self.species = species

        if self.nbodies == 1:
            for i in range(len(self.species)):
                isp = self.species[i]
                self.partial[isp] = corr_cls(*args, **kwargs)
                self.partial[isp].add_filter(filter_species, isp)
                self.partial[isp].tag = str(isp)
                self.partial[isp].tag_description = 'species %s' % isp

        elif self.nbodies == 2:
            for i in range(len(self.species)):
                for j in range(len(self.species)):
                    isp = self.species[i]
                    jsp = self.species[j]
                    self.partial[(isp, jsp)] = corr_cls(*args, **kwargs)
                    self.partial[(isp, jsp)].add_filter(filter_species, isp)
                    # Slight optimization: avoid filtering twice when isp==jsp
                    if isp != jsp:
                        self.partial[(isp, jsp)].add_filter(filter_species, jsp)
                    self.partial[(isp, jsp)].tag = '%s-%s' % (isp, jsp)
                    self.partial[(isp, jsp)].tag_description = 'species pair %s-%s' % (isp, jsp)

    def add_weight(self, trajectory=None, field=None, fluctuations=False):
        for key in self.partial:
            self.partial[key].add_weight(trajectory, field, fluctuations)
            
    def need_update(self):
        need = False
        for partial in self.partial.values():
            if partial.need_update():
                need = True
                break
        return need

    def compute(self):
        if self.nbodies == 1:
            for i in range(len(self.species)):
                isp = self.species[i]
                self.partial[isp].compute()

        elif self.nbodies == 2:
            for i in range(len(self.species)):
                for j in range(len(self.species)):
                    isp = self.species[i]
                    jsp = self.species[j]
                    if j >= i:
                        self.partial[(isp, jsp)].compute()
                    else:
                        # The isp-jsp has been already calculated
                        self.partial[(isp, jsp)].grid = self.partial[(jsp, isp)].grid
                        self.partial[(isp, jsp)].value = self.partial[(jsp, isp)].value

    def do(self, update=False):
        if update and not self.need_update():
            return

        self.compute()

        for partial in self.partial.values():
            try:
                partial.analyze()
            except ImportError as e:
                _log.warn('Could not analyze due to missing modules, continuing...')
                _log.warn(e.message)

            partial.write()



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Velocity autocorrelation function."""

import numpy

from .correlation import Correlation
from .progress import progress

__all__ = ['Fraction']


class Fraction(Correlation):

    """Fraction"""

    symbol = 'fraction'
    short_name = 'fraction()'
    long_name = 'fraction'
    phasespace = []

    def __init__(self, trajectory, condition=None, norigins=None):
        Correlation.__init__(self, trajectory, grid=[], norigins=norigins)
        self.condition = condition
        tag = condition.replace(' ', '')
        tag = tag.replace('and', '_')
        tag = tag.replace('or', '+')
        tag = tag.replace('"', '')
        tag = tag.replace("'", '')
        tag = tag.replace('==', '')
        tag = tag.replace('!=', '.not')
        tag = tag.replace(',', '-')
        self.tag = tag
        self.tag_description = condition

    def _compute(self):
        origins = range(0, len(self.trajectory), self.skip)
        result = 0.0
        norm = 0
        for i in progress(origins):
            norm += 1
            system = self.trajectory.read(i)
            cnt = 0
            for particle in system.particle:
                if eval(self.condition, particle.__dict__):
                    cnt += 1
            # TODO: fixme
            # result += float(cnt) /
        self.value = []
        self.grid = []
        self.analysis = {'fraction': result / float(norm)}



# This file is part of atooms
# Copyright 2010-2014, Daniele Coslovich

"""Linked cells to compute neighbors efficiently."""

from collections import defaultdict
import numpy

def _pbc(t, N):
    for i in range(len(t)):
        if t[i] >= N[i]:
            t[i] -= N[i]
        elif t[i] < 0:
            t[i] += N[i]
    return t

# TODO: define iterator over cells

class LinkedCells(object):

    def __init__(self, rcut):
        self.rcut = rcut
        self.neighbors = []
        self._is_adjusted = False

    def _map(self, newton):
        self._neigh_cell = {}
        if newton:
            for ix in range(self.n_cell[0]):
                for iy in range(self.n_cell[1]):
                    for iz in range(self.n_cell[2]):
                        self._neigh_cell[(ix, iy, iz)] = \
                            [(ix+1, iy, iz  ), (ix+1, iy+1, iz  ), (ix  , iy+1, iz  ), (ix-1, iy+1, iz  ),
                             (ix+1, iy, iz-1), (ix+1, iy+1, iz-1), (ix  , iy+1, iz-1), (ix-1, iy+1, iz-1),
                             (ix+1, iy, iz+1), (ix+1, iy+1, iz+1), (ix  , iy+1, iz+1), (ix-1, iy+1, iz+1),
                             (ix  , iy, iz+1)]
        else:
            for ix in range(self.n_cell[0]):
                for iy in range(self.n_cell[1]):
                    for iz in range(self.n_cell[2]):
                        self._neigh_cell[(ix, iy, iz)] = []
                        for deltax in [-1, 0, 1]:
                            for deltay in [-1, 0, 1]:
                                for deltaz in [-1, 0, 1]:
                                    if deltax == deltay == deltaz == 0:
                                        continue
                                    self._neigh_cell[(ix, iy, iz)].append((ix+deltax, iy+deltay, iz+deltaz))
                        
        # Apply PBC to neighboring "ghost" cells 
        for i in self._neigh_cell:
            for j in range(len(self._neigh_cell[i])):
                # if numpy.any(numpy.array(self._neigh_cell[i][j]) < 0) or \
                #    numpy.any(numpy.array(self._neigh_cell[i][j]) >= self.n_cell):
                folded = _pbc(list(self._neigh_cell[i][j]), self.n_cell)
                self._neigh_cell[i][j] = tuple(folded)

    def adjust(self, box, newton):
        self.box = box
        self.hbox = box / 2
        self.n_cell = (box / self.rcut).astype(int)
        self.box_cell = self.box / self.n_cell
        self._map(newton=newton)

    def _index(self, pos):
        x = ((pos + self.hbox) / self.box_cell)
        return x.astype(numpy.int)

    def on_border(self, pos):
        index = list(self._index(pos))
        found = False
        for i in range(len(index)):
            if index[i] == self.n_cell[i] - 1 or index[i] == 0:
                found = True
                break
        return found

    def compute(self, box, pos, other=None, as_array=False, newton=True):
        if not self._is_adjusted:
            self.adjust(box, newton=other is None and newton)
            self._is_adjusted = True
        # We only need positions here but how can we be sure that
        # this is the same set of particles we use when retrieving
        # the neighbours? We should keep a reference.
        # from atooms.postprocessing.realspace_wrap import compute
        # index = numpy.ndarray(pos.shape, dtype=numpy.int)
        # compute.bin_in_cell(pos, self.hbox, self.box_cell, index)

        self.neighbors = []
        self.number_of_neighbors = []
        index = self._index(pos)
        if other is None:
            index_other = index
        else:
            index_other = self._index(other)

        particle_in_cell = defaultdict(list)
        for ipart, icell in enumerate(index_other):
            particle_in_cell[tuple(icell)].append(ipart)

        for ipart in range(pos.shape[0]):
            icell = tuple(index[ipart])
            # Initialize an empty list
            neighbors = []
            # Start with particles in the cell of particle ipart
            if other is None:
                neighbors += [_ for _ in particle_in_cell[icell] if _ > ipart ]
            else:
                neighbors += particle_in_cell[icell]
                # try:
                #     neighbors.remove(ipart)
                # except:
                #     pass
            # Loop over neighbors cells and add neighbors
            for jcell in self._neigh_cell[icell]:
                neighbors += particle_in_cell[jcell]
            self.neighbors.append(neighbors)
            self.number_of_neighbors.append(len(neighbors))
            
        if as_array:
            npart = len(self.neighbors)
            number_of_neighbors = numpy.array(self.number_of_neighbors)
            neighbors_array = numpy.ndarray((npart, max(number_of_neighbors)), dtype=numpy.int)
            for ipart in range(len(self.neighbors)):
                neighbors_array[ipart, 0:len(self.neighbors[ipart])] = self.neighbors[ipart]
            self.neighbors = neighbors_array
            self.number_of_neighbors = number_of_neighbors
            return self.neighbors, number_of_neighbors
        else:            
            return self.neighbors
    
# import sys
# from atooms.system import Particle
# import atooms.trajectory as trj
# t = trj.Trajectory(sys.argv[1])
# system = t[0]
# # system.particle = []
# # dr = system.cell.side[0] / 20
# # for ix in range(0,20):
# #     for iy in range(0,20):
# #         system.particle.append(Particle(position=[ix*dr+dr/2 - system.cell.side[0]/2, iy*dr+dr/2 - system.cell.side[0]/2, 0]))

# trj.decorators.change_species(system, 'F')
# print('----')
# pos = system.dump('pos', order='C')
# ids = system.dump('spe')
# print('----')
# lc = _LinkedCells(rcut=2.0)
# nn, num = lc.compute(system.cell.side, pos, as_array=True)
# print(nn[0][0:num[0]])
# print('#', lc.box_cell / (system.cell.side/2))
# print('----')

# for j in nn[0][0:num[0]]:
#     dr = system.particle[0].distance(system.particle[j], system.cell)
#     print(numpy.sum(dr**2)**0.5, lc.box_cell[0] * 2)

# # from atooms.postprocessing.realspace_wrap import compute
# # rcut = numpy.array([[1.5, 1.5], [1.5, 1.5]])
# # rcut[:] = lc.box_cell[0] * 2
# # max_neighbors = 300
# # number_of_neighbors = numpy.ndarray(pos.shape[0], dtype=numpy.int32)
# # neighbors = numpy.ndarray((pos.shape[0], max_neighbors), dtype=numpy.int32, order='F')
# # #print(pos.shape)
# # pos = numpy.array(pos, order='F')
# # compute.neighbors_list('C',system.cell.side,pos.transpose(),ids,rcut,number_of_neighbors,neighbors)
# # #neighbors_0 = numpy.ndarray(max_neighbors, dtype=numpy.int32, order='F')
# # #nn = numpy.array(0, dtype=numpy.int32)
# # # compute.neighbors('C',system.cell.side,pos,pos[0],ids,rcut,nn,neighbors_0)
# # # print(neighbors[0][0:number_of_neighbors[0]])


# # print(str(system.particle[0].position / (system.cell.side/2))[1:-1])
# # print()
# # print()
# # for nnn in [sorted(neighbors[0]), sorted(lc.neighbors[0])]:
# #     for j in nnn:
# #         print(str(system.particle[j].position / (system.cell.side/2))[1:-1])
# #     print()
# #     print()



# # js = []
# # for j in range(pos.shape[0]):
# #     dr = system.particle[0].distance(system.particle[j], system.cell)
# #     if j> 0 and numpy.sum(dr**2)**0.5 < lc.box_cell[0] * 2:
# #         js.append(j)

# # # print(len(sorted(lc.neighbors[0])))
# # # print(len(sorted(neighbors[0][0:number_of_neighbors[0]])))
# # # print(len(sorted(js)))
# # # print('----')    

# # # for nnn in [sorted(lc.neighbors[0]),
# # #             sorted(neighbors[0][0:number_of_neighbors[0]])]:
# # #     for j in nnn:
# # #         print(str(system.particle[j].position)[1:-1])
# # #     print()
# # #     print()

# # # for nnn in [sorted(js),
# # #             sorted(neighbors[0][0:number_of_neighbors[0]])]:
# # #     for j in nnn:
# # #         dr = system.particle[0].distance(system.particle[j], system.cell)
# # #         print(j, numpy.sum(dr**2)**0.5)
# # #     print


# # # # for x in neighbors:
# # # #     print(x)
# # # for i in range(pos.shape[0]):
# # #     x = list(sorted(lc.neighbors[i]))
# # #     y = neighbors[i][0:number_of_neighbors[i]]
# # #     print(x)
# # #     print(y)
# # #     assert numpy.all(x == y)
# # #     print()



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Non-Gaussian parameter."""

import numpy

from .helpers import linear_grid
from .correlation import Correlation, gcf_offset
from .helpers import setup_t_grid, ifabsmm

__all__ = ['NonGaussianParameter']


class NonGaussianParameter(Correlation):

    """Non-Gaussian parameter."""

    symbol = 'alpha2'
    short_name = 'alpha_2(t)'
    long_name = 'non-Gaussian parameter'
    phasespace = ['pos-unf']

    def __init__(self, trajectory, tgrid=None, norigins=None, nsamples=30):
        Correlation.__init__(self, trajectory, tgrid, norigins=norigins)
        if self.grid is None:
            self.grid = linear_grid(0.0, self.trajectory.total_time * 0.75, nsamples)
        self._discrete_tgrid = setup_t_grid(self.trajectory, self.grid, offset=norigins != '1')

    def _compute(self):
        def alpha_2(x, y):
            if x is y:
                return 0.0
            dx2 = (x - y)**2
            dr2 = numpy.sum(dx2) / float(x.shape[0])
            dr4 = numpy.sum(numpy.sum(dx2, axis=1)**2) / float(x.shape[0])
            return 3 * dr4 / (5 * dr2**2) - 1

        self.grid, self.value = gcf_offset(alpha_2, self._discrete_tgrid, self.skip,
                                           self.trajectory.steps, self._pos_unf)
        self.grid = [ti * self.trajectory.timestep for ti in self.grid]

    def analyze(self):
        try:
            self.analysis['t_star'], self.analysis['a2_star'] = ifabsmm(self.grid, self.value)[1]
        except ZeroDivisionError:
            pass



#!/usr/bin/env python
# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Stress autocorrelation function."""

import numpy

from .correlation import Correlation, gcf_offset
from .helpers import setup_t_grid

__all__ = ['StressAutocorrelation']


class StressAutocorrelation(Correlation):

    """Stress autocorrelation function."""

    symbol = 'sacf'
    short_name = 'S(t)'
    long_name = 'stress autocorrelation'
    phasespace = ['vel']

    def __init__(self, trajectory, tgrid, norigins=None):
        Correlation.__init__(self, trajectory, tgrid, norigins=norigins)
        self._discrete_tgrid = setup_t_grid(self.trajectory, tgrid, offset=norigins != '1')

    def _get_stress(self):
        ndims = 3
        p = self.trajectory.read(0).particle
        mass = numpy.array([pi.mass for pi in p])
        self._stress = []
        for i in self.trajectory.samples:
            s = self.trajectory.read(i).interaction.stress
            slk = numpy.zeros(ndims)
            l = 0
            for j in range(ndims):
                for k in range(j+1, ndims):
                    slk[l] = s[j, k] + numpy.sum(mass[:] * self._vel[i][:, j] * self._vel[i][:, k])
                    l += 1
            self._stress.append(slk)

    def _compute(self):
        def f(x, y):
            return numpy.sum(x*y) / float(x.shape[0])

        self._get_stress()
        V = self.trajectory.read(0).cell.volume
        self.grid, self.value = gcf_offset(f, self._discrete_tgrid, self.trajectory.block_size,
                                           self.trajectory.steps, self._stress)
        self.value = [x / V for x in self.value]
        self.grid = [ti * self.trajectory.timestep for ti in self.grid]

    def analyze(self):
        pass



import copy


def linear_grid(min_val, max_val, delta):
    """Linear grid."""
    if isinstance(delta, int):
        n = delta
        if n > 1:
            delta = (max_val - min_val) / (n-1)
        else:
            delta = 0.0
    else:
        n = int((max_val - min_val) / delta) + 1
    return [min_val + i*delta for i in range(n)]


def logx_grid(x1, x2, n):
    """Create a list of n numbers in logx scale from x1 to x2."""
    # the shape if a*x^n. if n=0 => a=x1, if n=N => x1*x^N=x2
    if x1 > 0:
        xx = (x2 / x1)**(1.0 / n)
        return [x1] + [x1 * xx**(i+1) for i in range(1, n)]
    else:
        xx = x2**(1.0/n)
        return [x1] + [xx**(i+1) - 1 for i in range(1, n)]


def ifabsmm(x, f):
    """
    Find absolute maximum and absolute minimum of f(x) using a
    parabolic interpolation.

    Given x and f(x), the function returns a tuple of two entries, the
    first one being (x_min, f(x_min)), the second (x_max, f(x_max)).

    If the maximum of the minimum lies on the boundary of the
    interval, no interpolation is performed.
    """

    def _vertex_parabola(a, b, c):
        """Returns the vertex (x,y) of a parabola of the type a*x**2 + b*x + c."""
        return -b/(2*a), - (b**2 - 4*a*c) / (4*a)

    def _parabola_3points(x1, y1, x2, y2, x3, y3):
        """Parabola through 3 points."""
        delta = (x1 - x2)*(x1 - x3)*(x2 - x3)
        a = (x3 * (y2 - y1) + x2 * (y1 - y3) + x1 * (y3 - y2)) / delta
        b = (x3**2 * (y1 - y2) + x2**2 * (y3 - y1) + x1**2 * (y2 - y3)) / delta
        c = (x2 * x3 * (x2 - x3) * y1 + x3 * x1 * (x3 - x1) * y2 + x1 * x2 * (x1 - x2) * y3) / delta
        return a, b, c

    # First uninterpolated minima and maxima
    f = list(f)
    imin, imax = f.index(min(f)), f.index(max(f))
    # Then perform parabolic interpolation
    ii = []
    for i in [imin, imax]:
        if i == len(f)-1 or i == 0:
            # At boundaries we do nothing
            ii.append((x[i], f[i]))
        else:
            # Parabola around i
            i1, i2, i3 = i-1, i, i+1
            a, b, c = _parabola_3points(x[i1], f[i1], x[i2], f[i2], x[i3], f[i3])
            ii.append(_vertex_parabola(a, b, c))
    return ii[0], ii[1]


def linear_fit(xdata, ydata):
    """
    Linear regression.

    Expressions as in Wikipedia (https://en.wikipedia.org/wiki/Simple_linear_regression)
    """
    import numpy
    from math import sqrt

    n = len(ydata)
    dof = n - 2
    sx = numpy.sum(xdata)
    sy = numpy.sum(ydata)
    sxy = sum(xdata * ydata)
    sxx = sum(xdata**2)
    syy = sum(ydata**2)

    a = (n * sxy - sx * sy) / (n * sxx - sx**2)
    b = sy / n - a * sx / n
    if dof > 0:
        s = (n*syy - sy**2 - a**2 * (n*sxx - sx**2)) / (n*dof)
    else:
        s = 0.0
    sa = n * s / (n*sxx - sx**2)
    sb = sa * sxx / n
    return a, b, sqrt(sa), sqrt(sb)


def feqc(x, f, fstar):
    """
    Find first root of f=f(x) for data sets.

    Given two lists x and f, it returns the value of xstar for which
    f(xstar) = fstar. Raises an ValueError if no root is found.
    """
    s = f[0] - fstar
    for i in range(min(len(x), len(f))):
        if (f[i] - fstar) * s < 0.0:
            # Linear interpolation
            dxf = (f[i] - f[i-1]) / (x[i] - x[i-1])
            xstar = x[i-1] + (fstar - f[i-1]) / dxf
            istar = i
            return xstar, istar

    # We get to the end and cannot find the root
    return None, None


def filter_species(system, species):
    """Callback to filter particles by species.

    The input species can be an integer (particle id), a string
    (particle name), or None. In this latter case, all particles
    are returned.
    """
    s = copy.copy(system)
    if species is not None:
        s.particle = [p for p in system.particle if p.species == species]
    return s


def copy_field(system, field, trajectory):
    """
    Copy particle property `field` from `trajectory` at the current
    frame in system.

    It requires atooms >= 1.10.0
    """
    # Only available in atooms > 1.10.0
    so = trajectory[system.frame]
    for p, po in zip(system.particle, so.particle):
        x = getattr(po, field)
        setattr(p, field, x)
    return system


def adjust_skip(trajectory, n_origins=None):
    """
    Define interval between frames in trajectory so as to achieve a
    given number of time origins.

    Possible values of n_origins:
    - None: heuristics to keep the product of steps * particles constant
    - int: if -1, all origins are used, otherwise if n_origins >= 1 only n_origins
    - float in the interval (0,1): the fraction of samples to consider as time origins
    """
    if n_origins is not None:
        if float(n_origins) < 0 or n_origins == '1.0':
            skip = 1 * trajectory.block_size  # all origins
        elif float(n_origins) > 1:
            nbl = len(trajectory.steps) // trajectory.block_size
            skip = int(nbl / float(n_origins)) * trajectory.block_size
        elif int(n_origins) == 1:
            skip = len(trajectory.steps)
        else:
            # A float between 0 and 1
            skip = int(1 / float(n_origins)) * trajectory.block_size
    else:
        # Heuristics (to be improved)
        if trajectory.block_size == 1:
            block = 40000
            # Ignore cache here
            # This destroys the connection between decorated unfolded trajectory and folded trajectory
            #_cache, trajectory.cache = trajectory.cache, False
            skip = int(len(trajectory.steps) * len(trajectory[0].particle) / block)
            #trajectory.cache = _cache
        else:
            # We stick to the block size
            skip = trajectory.block_size

    # Make sure the skip is an multiple of block size
    if skip % trajectory.block_size != 0 and int(n_origins) != 1:
        raise ValueError('wrong skip {} {}'.format(skip, trajectory.block_size))

    # Normalize anyway and make it even
    skip = max(1, skip)
    skip = min(len(trajectory.steps), skip)
    return skip


def _templated(entry, template, keep_multiple=False):
    """
    Filter a list of entries so as to best match an input
    template. Lazy, slow version O(N*M).

    Example:
    --------
    entry = [1,2,3,4,5,10,20,100]
    template = [1,7,12,80]
    templated(entry, template) == [1,5,10,100]
    """
    match = []
    for t in template:
        def compare(x):
            return abs(x - t)
        match.append(min(entry, key=compare))
    if not keep_multiple:
        match = list(set(match))
    return sorted(match)


def setup_t_grid(trajectory, t_grid, offset=True):
    # First get all possible time differences
    steps = trajectory.steps
    off_samp = {}
    if offset:
        offset_range = range(trajectory.block_size)
    else:
        offset_range = [0]
    for off in offset_range:
        for i in range(off, len(steps)-off):
            if steps[i] - steps[off] not in off_samp:
                off_samp[steps[i] - steps[off]] = (off, i-off)

    # Retain only those pairs of offsets and sample
    # difference that match the desired input. This is the grid
    # used internally to calculate the time correlation function.
    i_grid = set([int(round(t/trajectory.timestep)) for t in t_grid])
    offsets = [off_samp[t] for t in _templated(sorted(off_samp.keys()), sorted(i_grid))]
    return offsets


def _dump(title, columns=None, command=None, version=None,
          description=None, note=None, parents=None, inline=False,
          comment='# ', extra_fields=None):
    """
    Return a string of comments filled with metadata.
    """
    import datetime

    date = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    if columns is not None:
        columns_string = ', '.join(columns)

    metadata = [('title', title),
                ('columns', columns_string),
                ('date', date),
                ('command', command),
                ('version', version),
                ('parents', parents),
                ('description', description),
                ('note', note)]

    if extra_fields is not None:
        metadata += extra_fields

    if inline:
        fmt = '{}: {};'
        txt = comment + ' '.join([fmt.format(key, value) for key,
                                  value in metadata if value is not None])
    else:
        txt = ''
        for key, value in metadata:
            if value is not None:
                txt += comment + '{}: {}\n'.format(key, value)
    return txt



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Time-dependent overlap."""

import numpy

from .helpers import logx_grid
from .correlation import Correlation, gcf_offset
from .helpers import setup_t_grid

__all__ = ['CollectiveOverlap', 'SelfOverlap']


def pairs_numpy(f, x, y, L):
    """
    Apply function f to all pairs in x[i] and y[j] and return a numpy
    array of f values.
    """
    fxy = numpy.ndarray((len(y), len(x)))
    for i in range(fxy.shape[0]):
        fxy[i, :] = f(x[:], y[i], L)
    return fxy

def square_displacement(x, y, L=None):
    """Return array of square distances (no pbc)."""
    return numpy.sum((x-y)**2, axis=1)

def collective_overlap(r0, r1, side, a_square):
    rij = pairs_numpy(square_displacement, r0, r1, side)
    return (rij.flatten() < a_square).sum()

def self_overlap(r0, r1, side, a_square):
    rij = square_displacement(r0, r1)
    return rij.flatten() < a_square


class CollectiveOverlap(Correlation):

    """Time-dependent collective overlap."""

    symbol = 'qt'
    short_name = 'Q(t)'
    long_name = 'collective overlap'
    phasespace = 'pos'

    def __init__(self, trajectory, tgrid=None, tsamples=60, a=0.3,
                 norigins=-1):
        Correlation.__init__(self, trajectory, tgrid, norigins=norigins)
        self.a_square = a**2
        if tgrid is None:
            self.grid = logx_grid(0.0, self.trajectory.total_time * 0.75, tsamples)
        self._discrete_tgrid = setup_t_grid(self.trajectory, self.grid, offset=norigins != '1')

    def _compute(self):
        side = self.trajectory.read(0).cell.side
        def f(x, y):
            return collective_overlap(x, y, side, self.a_square).sum() / float(x.shape[0])
        self.grid, self.value = gcf_offset(f, self._discrete_tgrid,
                                           self.skip, self.trajectory.steps, self._pos)
        self.grid = [ti * self.trajectory.timestep for ti in self.grid]


class SelfOverlap(Correlation):

    """Time-dependent self overlap."""

    symbol = 'qst'
    short_name = 'Q_s(t)'
    long_name = 'self overlap'
    phasespace = 'pos-unf'

    def __init__(self, trajectory, tgrid=None, norigins=-1, a=0.3,
                 tsamples=60):
        Correlation.__init__(self, trajectory, tgrid, norigins=norigins)
        if tgrid is None:
            self.grid = logx_grid(0.0, trajectory.total_time * 0.75, tsamples)
        self._discrete_tgrid = setup_t_grid(self.trajectory, self.grid, offset=norigins != '1')
        self.a_square = a**2

    def _compute(self):
        side = self.trajectory.read(0).cell.side
        def f(x, y):
            return self_overlap(x, y, side, self.a_square).sum() / float(x.shape[0])
        self.grid, self.value = gcf_offset(f, self._discrete_tgrid,
                                           self.skip, self.trajectory.steps, self._pos_unf)
        self.grid = [ti * self.trajectory.timestep for ti in self.grid]

    def analyze(self):
        try:
            from .helpers import feqc
            self.analysis['tau'] = feqc(self.grid, self.value, 1 / numpy.exp(1.0))[0]
        except ValueError:
            self.analysis['tau'] = None



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Structure factor."""

import logging
import numpy

from .progress import progress
from .fourierspace import FourierSpaceCorrelation, expo_sphere

__all__ = ['StructureFactor', 'StructureFactorLegacy', 'StructureFactorOptimized', 'StructureFactorFast']

_log = logging.getLogger(__name__)


def is_cell_variable(trajectory, tests=1):
    """
    Simple test to check if cell changes.

    We compare the first frame to an integer number `tests` of other
    frames starting from the end of `trajectory`.
    """
    is_variable = False
    frames = len(trajectory)
    if tests > 0:
        skip = max(1, int(frames / float(tests)))
    else:
        skip = 1
    L0 = trajectory[0].cell.side
    for sample in range(frames-1, 0, -skip):
        cell = trajectory[sample].cell
        L1 = cell.side
        if numpy.any(L0 != L1):
            is_variable = True
            break
    return is_variable


class StructureFactorLegacy(FourierSpaceCorrelation):
    """
    Structure factor.

    See the documentation of the `FourierSpaceCorrelation` base class
    for information on the instance variables.
    """

    nbodies = 2
    symbol = 'sk'
    short_name = 'S(k)'
    long_name = 'structure factor'
    phasespace = ['pos']

    def __init__(self, trajectory, kgrid=None, norigins=-1, nk=20,
                 dk=0.1, kmin=-1.0, kmax=15.0, ksamples=30):
        FourierSpaceCorrelation.__init__(self, trajectory, kgrid, norigins,
                                         nk, dk, kmin, kmax, ksamples)
        self._is_cell_variable = None

    def _compute(self):
        nsteps = len(self._pos_0)
        ndims = len(self.k0)
        # Setup k vectors and tabulate rho
        kgrid, selection = self.kgrid, self.selection
        kmax = max(self.kvector.keys()) + self.dk
        cnt = [0 for k in kgrid]
        rho_0_av = [complex(0., 0.) for k in kgrid]
        rho_1_av = [complex(0., 0.) for k in kgrid]
        rho2_av = [complex(0., 0.) for k in kgrid]
        variable_cell = is_cell_variable(self.trajectory)

        for i in progress(range(0, nsteps, self.skip), total=nsteps // self.skip):
            # If cell changes we have to update the wave vectors
            if variable_cell:
                self._setup(i)
                kgrid, selection = self._decimate_k()
                kmax = max(self.kvector.keys()) + self.dk

            # Tabulate exponentials
            # Note: tabulating and computing takes about the same time
            if self._pos_0[i] is self._pos_1[i]:
                # Identical species
                expo_0 = expo_sphere(self.k0, kmax, self._pos_0[i])
                expo_1 = expo_0
            else:
                # Cross correlation
                expo_0 = expo_sphere(self.k0, kmax, self._pos_0[i])
                expo_1 = expo_sphere(self.k0, kmax, self._pos_1[i])

            # Define weights
            if self._weight is None:
                weight_0, weight_1 = 1.0, 1.0
            else:
                weight_0, weight_1 = self._weight_0[i], self._weight_1[i]

            # Nice spaghetti here
            for kk, knorm in enumerate(kgrid):
                for k in selection[kk]:
                    ik = self.kvector[knorm][k]
                    if expo_0 is expo_1:
                        # Identical species
                        if ndims == 3:
                            rho_0 = numpy.sum(weight_0 *
                                              expo_0[..., 0, ik[0]] *
                                              expo_0[..., 1, ik[1]] *
                                              expo_0[..., 2, ik[2]])
                        elif ndims == 2:
                            rho_0 = numpy.sum(weight_0 *
                                              expo_0[..., 0, ik[0]] *
                                              expo_0[..., 1, ik[1]])
                        else:
                            tmp = weight_0 * expo_0[..., 0, ik[0]]
                            for idim in range(1, ndims):
                                tmp *= expo_0[..., idim, ik[idim]]
                            rho_0 = numpy.sum(tmp)

                        rho_1 = rho_0
                    else:
                        # Cross correlation
                        if ndims == 3:
                            rho_0 = numpy.sum(weight_0 *
                                              expo_0[..., 0, ik[0]] *
                                              expo_0[..., 1, ik[1]] *
                                              expo_0[..., 2, ik[2]])
                            rho_1 = numpy.sum(weight_1 *
                                              expo_1[..., 0, ik[0]] *
                                              expo_1[..., 1, ik[1]] *
                                              expo_1[..., 2, ik[2]])
                        elif ndims == 2:
                            rho_0 = numpy.sum(weight_0 *
                                              expo_0[..., 0, ik[0]] *
                                              expo_0[..., 1, ik[1]])
                            rho_1 = numpy.sum(weight_1 *
                                              expo_1[..., 0, ik[0]] *
                                              expo_1[..., 1, ik[1]])
                        else:
                            tmp = weight_0 * expo_0[..., 0, ik[0]]
                            for idim in range(ndims):
                                tmp *= expo_0[..., idim, ik[idim]]
                            rho_0 = numpy.sum(tmp)
                            tmp = weight_0 * expo_1[..., 0, ik[0]]
                            for idim in range(ndims):
                                tmp *= expo_1[..., idim, ik[idim]]
                            rho_1 = numpy.sum(tmp)
                            
                    # Cumulate averages
                    rho_0_av[kk] += rho_0
                    rho_1_av[kk] += rho_1
                    rho2_av[kk] += (rho_0 * rho_1.conjugate())
                    cnt[kk] += 1

        # In the absence of a microscopic field, the average density is zero.
        # We get rid of the average and compute <rho(k)rho*(k)>.
        if self._weight_0 is None:
            rho_0_av = [complex(0., 0.) for k in kgrid]
            rho_1_av = [complex(0., 0.) for k in kgrid]

        # Normalization
        npart_0 = sum([p.shape[0] for p in self._pos_0]) / float(len(self._pos_0))
        npart_1 = sum([p.shape[0] for p in self._pos_1]) / float(len(self._pos_1))
        self.grid = kgrid
        self.value, self.value_nonorm = [], []
        for kk in range(len(self.grid)):
            norm = float(npart_0 * npart_1)**0.5
            value = (rho2_av[kk] / cnt[kk] - rho_0_av[kk]*rho_1_av[kk].conjugate() / cnt[kk]**2).real
            self.value.append(value / norm)
            self.value_nonorm.append(value)


class StructureFactorFast(StructureFactorLegacy):
    """
    Optimized structure factor.

    It uses a fortran 90 extension.
    """

    nbodies = 2
    symbol = 'sk'
    short_name = 'S(k)'
    long_name = 'structure factor'
    phasespace = ['pos']

    def _compute(self):
        from atooms.trajectory.utils import is_cell_variable
        try:
            from atooms.postprocessing.fourierspace_wrap import fourierspace_module
        except ImportError:
            _log.error('f90 wrapper missing or not functioning')
            raise

        nsteps = len(self._pos_0)
        # Setup k vectors and tabulate rho
        kgrid, selection = self.kgrid, self.selection
        kmax = max(self.kvector.keys()) + self.dk
        cnt = [0 for k in kgrid]
        rho_av = [complex(0., 0.) for k in kgrid]
        rho2_av = [complex(0., 0.) for k in kgrid]
        variable_cell = is_cell_variable(self.trajectory)
        for i in range(0, nsteps, self.skip):
            # If cell changes we have to update the wave vectors
            if variable_cell:
                self._setup(i)
                kgrid, selection = self._decimate_k()
                kmax = max(self.kvector.keys()) + self.dk

            # Tabulate exponentials
            # Note: tabulating and computing takes about the same time
            if self._pos_0[i] is self._pos_1[i]:
                # Identical species
                expo_0 = expo_sphere(self.k0, kmax, self._pos_0[i])
                expo_1 = expo_0
            else:
                # Cross correlation
                # TODO: cross correlation wont work
                expo_0 = expo_sphere(self.k0, kmax, self._pos_0[i])
                expo_1 = expo_sphere(self.k0, kmax, self._pos_1[i])

            for kk, knorm in enumerate(kgrid):
                ikvec = numpy.ndarray((3, len(selection[kk])), order='F', dtype=numpy.int32)
                i = 0
                for k in selection[kk]:
                    ikvec[:, i] = self.kvector[knorm][k]
                    i += 1
                rho = numpy.zeros(ikvec.shape[1], dtype=numpy.complex128)
                fourierspace_module.sk_bare(expo_0, ikvec, rho)
                rho_0 = rho
                rho_1 = rho
                rho2_av[kk] += numpy.sum(rho_0 * rho_1.conjugate())
                cnt[kk] += rho.shape[0]

        # Normalization.
        npart_0 = sum([p.shape[0] for p in self._pos_0]) / float(len(self._pos_0))
        npart_1 = sum([p.shape[0] for p in self._pos_1]) / float(len(self._pos_1))
        self.grid = kgrid
        self.value, self.value_nonorm = [], []
        for kk in range(len(self.grid)):
            norm = float(npart_0 * npart_1)**0.5
            value = (rho2_av[kk] / cnt[kk] - rho_av[kk]*rho_av[kk].conjugate() / cnt[kk]**2).real
            self.value.append(value / norm)
            self.value_nonorm.append(value)


# Defaults to legacy
StructureFactor = StructureFactorLegacy
# Backward compatible alias
StructureFactorOptimized = StructureFactorFast



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Intermediate scattering function."""

import logging
from collections import defaultdict

import numpy
from atooms.trajectory.utils import check_block_size

from . import core
from atooms.trajectory import Trajectory
from .helpers import logx_grid, setup_t_grid
from .correlation import Correlation
from .fourierspace import FourierSpaceCorrelation, expo_sphere
from .progress import progress

__all__ = ['SelfIntermediateScattering',
           'SelfIntermediateScatteringLegacy', 'SelfIntermediateScatteringFast',
           'IntermediateScattering']

_log = logging.getLogger(__name__)


def _write_tau(out, db):
    # Custom writing of relaxation times
    out.write('# title: relaxation times tau(k) as a function of k\n')
    out.write('# columns: k, tau(k)\n')
    out.write('# note: tau is the time at which the correlation function has decayed to 1/e\n')
    for k, tau in db['relaxation times tau'].items():
        if tau is None:
            out.write('%g\n' % k)
        else:
            out.write('%g %g\n' % (k, tau))

def _extract_tau(k, t, f, factor=1/numpy.exp(1.0)):
    from .helpers import feqc
    # Ensure first point in time grid is t=0
    # This will be enforced by the classes below but we never know
    if t[0] > 0:
        raise ValueError('First point in time grid must be zero')
    tau = {}
    for i, k in enumerate(k):
        try:
            tau[k] = feqc(t, f[i], f[i][0]*factor)[0]
        except ValueError:
            tau[k] = None
    return tau


class IntermediateScatteringBase(FourierSpaceCorrelation):

    def __init__(self, trajectory, kgrid=None, tgrid=None, nk=1,
                 tsamples=1, dk=0.1, kmin=1.0, kmax=10.0, ksamples=10,
                 norigins=-1, fix_cm=False, normalize=True):
        FourierSpaceCorrelation.__init__(self, trajectory, [kgrid, tgrid], norigins,
                                         nk, dk, kmin, kmax, ksamples, fix_cm, normalize)
        # Before setting up the time grid, we need to check periodicity over blocks
        try:
            check_block_size(self.trajectory.steps, self.trajectory.block_size)
        except IndexError as e:
            _log.warn('issue with trajectory blocks, the time grid may not correspond to the requested one (%s)', e.message)

        # Setup time grid
        if self.grid[1] is None:
            self.grid[1] = logx_grid(0.0, self.trajectory.total_time * 0.75, tsamples)
        else:
            # If the values are normalized, we make sure the
            # user-provided time grid includes t=0. It is removed
            # after normalization
            if self.normalize:
                if self.grid[1][0] > 0:
                    _log.info('adding t=0 to the time grid to normalize F_s(k,t)')
                    self.grid[1] = [0.0] + self.grid[1]

        # When a single time origin is requested,
        # make sure no additional time origins except the first frame is used
        self._discrete_tgrid = setup_t_grid(self.trajectory, self.grid[1], offset=norigins != '1' and norigins != 1)


class SelfIntermediateScatteringLegacy(IntermediateScatteringBase):
    """
    Self part of the intermediate scattering function.

    See the documentation of the `FourierSpaceCorrelation` base class
    for information on the instance variables.
    """

    symbol = 'fskt'
    short_name = 'F_s(k,t)'
    long_name = 'self intermediate scattering function'
    phasespace = 'pos'

    def __init__(self, trajectory, kgrid=None, tgrid=None, nk=8,
                 tsamples=60, dk=0.1, kmin=1.0, kmax=10.0,
                 ksamples=10, norigins=-1, fix_cm=False,
                 lookup_mb=64.0, normalize=True):
        # TODO: remove this backward compatible tgrid fix in a major release
        # The default time grid should be the same for F_s(k,t) and F(k,t)
        if isinstance(trajectory, str):
            trajectory = Trajectory(trajectory, mode='r', fmt=core.pp_trajectory_format)
        if tgrid is None:
            tgrid = [0.0] + logx_grid(trajectory.timestep,
                                      trajectory.total_time * 0.75, tsamples)
        super(SelfIntermediateScatteringLegacy,
              self).__init__(trajectory, kgrid=kgrid, tgrid=tgrid,
                             nk=nk, tsamples=tsamples, dk=dk, kmin=kmin,
                             kmax=kmax, ksamples=ksamples, norigins=norigins,
                             fix_cm=fix_cm, normalize=normalize)
        self.lookup_mb = lookup_mb
        """Memory in Mb allocated for exponentials tabulation"""

    def _compute(self):
        # Throw everything into a big numpy array (nframes, npos, ndim)
        pos = numpy.array(self._pos)
        ndims = len(self.k0)
        # To optimize without wasting too much memory (we have
        # troubles here) we group particles in blocks and tabulate the
        # exponentials over time. This is more memory consuming but we
        # can optimize the inner loop. Even better, we could change
        # the order in the tabulated expo array to speed things up
        # Use 10 blocks, but do not exceed 200 particles
        number_of_blocks = 10
        block = int(pos[0].shape[0] / float(number_of_blocks))
        block = max(20, block)
        block = min(200, block)
        if len(self.kvector.keys()) == 0:
            raise ValueError('could not find any wave-vectors, try increasing dk')
        kmax = max(self.kvector.keys()) + self.dk
        acf = [defaultdict(float) for _ in self.kgrid]
        cnt = [defaultdict(float) for _ in self.kgrid]
        skip = self.skip
        origins = range(0, pos.shape[1], block)
        for j in progress(origins):
            x = expo_sphere(self.k0, kmax, pos[:, j:j + block, :])
            for kk, knorm in enumerate(self.kgrid):
                for kkk in self.selection[kk]:
                    ik = self.kvector[knorm][kkk]
                    for off, i in self._discrete_tgrid:
                        for i0 in range(off, x.shape[0]-i, skip):
                            # Get the actual time difference. steps must be accessed efficiently (cached!)
                            dt = self.trajectory.steps[i0+i] - self.trajectory.steps[i0]
                            # Dimensional switch
                            if ndims == 3:
                                acf[kk][dt] += numpy.sum(x[i0+i, :, 0, ik[0]]*x[i0, :, 0, ik[0]].conjugate() *
                                                         x[i0+i, :, 1, ik[1]]*x[i0, :, 1, ik[1]].conjugate() *
                                                         x[i0+i, :, 2, ik[2]]*x[i0, :, 2, ik[2]].conjugate()).real
                            elif ndims == 2:
                                acf[kk][dt] += numpy.sum(x[i0+i, :, 0, ik[0]]*x[i0, :, 0, ik[0]].conjugate() *
                                                         x[i0+i, :, 1, ik[1]]*x[i0, :, 1, ik[1]].conjugate()).real

                            else:
                                # Arbitrary dimension (a bit slower)
                                tmp = x[i0+i, :, 0, ik[0]]*x[i0, :, 0, ik[0]].conjugate()
                                for idim in range(1, len(ik)):
                                    tmp *= x[i0+i, :, idim, ik[idim]]*x[i0, :, idim, ik[idim]].conjugate()
                                acf[kk][dt] += numpy.sum(tmp).real                            
                            cnt[kk][dt] += x.shape[1]

        tgrid = sorted(acf[0].keys())
        self.grid[0] = self.kgrid
        self.grid[1] = [ti*self.trajectory.timestep for ti in tgrid]
        self.value = [[acf[kk][ti] / cnt[kk][ti] for ti in tgrid] for kk in range(len(self.grid[0]))]

        # Normalize
        if self.normalize:
            for k in range(len(self.grid[0])):
                for i in range(len(self.value[k])):
                    self.value[k][i] /= self.value[k][0]

    def analyze(self):
        self.analysis['relaxation times tau'] = _extract_tau(self.grid[0], self.grid[1], self.value)

    def write(self):
        Correlation.write(self)
        if self._output_file != '/dev/stdout':
            with open(self._output_file + '.tau', 'w') as out:
                _write_tau(out, self.analysis)


class SelfIntermediateScatteringFast(SelfIntermediateScatteringLegacy):
    """
    Self part of the intermediate scattering function (fast version)
    
    See the documentation of the `FourierSpaceCorrelation` base class
    for information on the instance variables.
    """        
    def _compute(self):
        from atooms.postprocessing.fourierspace_wrap import fourierspace_module

        # Throw everything into a big numpy array (nframes, npos, ndim)
        pos = numpy.array(self._pos)

        # Select the f90 kernel
        ndims = len(self.k0)
        if ndims == 3:
            fskt_kernel = fourierspace_module.fskt_kernel_3d
        elif ndims == 2:
            fskt_kernel = fourierspace_module.fskt_kernel_2d
        else:
            fskt_kernel = fourierspace_module.fskt_kernel_nd
            
        # To optimize without wasting too much memory (we have
        # troubles here) we group particles in blocks and tabulate the
        # exponentials over time. This is more memory consuming but we
        # can optimize the inner loop. The esitmated amuount of
        # allocated memory in Mb for the expo array is
        # self.lookup_mb. Note that the actual memory need scales
        # with number of k vectors, system size and number of frames.
        kmax = max(self.kvector.keys()) + self.dk
        kvec_size = 2*(1 + int(kmax / min(self.k0))) + 1
        pos_size = numpy.product(pos.shape)
        target_size = self.lookup_mb * 1e6 / 16.  # 16 bytes for a (double) complex        
        number_of_blocks = int(pos_size * kvec_size / target_size)
        number_of_blocks = max(1, number_of_blocks)
        block = int(pos[0].shape[0] / float(number_of_blocks))
        block = max(1, block)
        block = min(pos.shape[1], block)
        if len(self.kvector.keys()) == 0:
            raise ValueError('could not find any wave-vectors, try increasing dk')
        acf = [defaultdict(float) for _ in self.kgrid]
        cnt = [defaultdict(float) for _ in self.kgrid]
        skip = self.skip
        origins = range(0, pos.shape[1], block)
        for j in progress(origins):
            x = expo_sphere(self.k0, kmax, pos[:, j:j + block, :])            
            xf = numpy.asfortranarray(x)
            for kk, knorm in enumerate(self.kgrid):
                for kkk in self.selection[kk]:
                    ik = self.kvector[knorm][kkk]
                    for off, i in self._discrete_tgrid:
                        for i0 in range(off, x.shape[0]-i, skip):
                            # Get the actual time difference. steps must be accessed efficiently (cached!)
                            dt = self.trajectory.steps[i0+i] - self.trajectory.steps[i0]
                            # Call f90 kernel
                            res = fskt_kernel(xf, i0+1, i0+1+i, numpy.array(ik, dtype=numpy.int32)+1)
                            acf[kk][dt] += res.real
                            cnt[kk][dt] += x.shape[1]
                            
        tgrid = sorted(acf[0].keys())
        self.grid[0] = self.kgrid
        self.grid[1] = [ti*self.trajectory.timestep for ti in tgrid]
        self.value = [[acf[kk][ti] / cnt[kk][ti] for ti in tgrid] for kk in range(len(self.grid[0]))]
        # Normalize
        if self.normalize:
            for k in range(len(self.grid[0])):
                for i in range(len(self.value[k])):
                    self.value[k][i] /= self.value[k][0]


# Defaults to fast
try:
    import atooms.postprocessing.fourierspace_wrap
    SelfIntermediateScattering = SelfIntermediateScatteringFast
except ImportError:
    SelfIntermediateScattering = SelfIntermediateScatteringLegacy


class IntermediateScattering(IntermediateScatteringBase):
    """
    Coherent intermediate scattering function.

    See the documentation of the `FourierSpaceCorrelation` base class
    for information on the instance variables.
    """

    nbodies = 2
    symbol = 'fkt'
    short_name = 'F(k,t)'
    long_name = 'intermediate scattering function'
    phasespace = 'pos'

    def __init__(self, trajectory, kgrid=None, tgrid=None, nk=100, dk=0.1, tsamples=60,
                 kmin=1.0, kmax=10.0, ksamples=10, norigins=-1, fix_cm=False, normalize=True):
        super(IntermediateScattering, self).__init__(trajectory, kgrid=kgrid, tgrid=tgrid,
                             nk=nk, tsamples=tsamples, dk=dk, kmin=kmin,
                             kmax=kmax, ksamples=ksamples, norigins=norigins,
                             fix_cm=fix_cm, normalize=normalize)

    def _tabulate_rho(self, kgrid, selection):
        """
        Tabulate densities
        """
        nsteps = len(self._pos_0)
        ndims = len(self.k0)
        if len(self.kvector.keys()) == 0:
            raise ValueError('could not find any wave-vectors, try increasing dk')
        kmax = max(self.kvector.keys()) + self.dk
        rho_0 = [defaultdict(complex) for it in range(nsteps)]
        rho_1 = [defaultdict(complex) for it in range(nsteps)]
        for it in range(nsteps):
            expo_0 = expo_sphere(self.k0, kmax, self._pos_0[it])
            # Optimize a bit here: if there is only one filter (alpha-alpha or total calculation)
            # expo_2 will be just a reference to expo_1
            if self._pos_1 is self._pos_0:
                expo_1 = expo_0
            else:
                expo_1 = expo_sphere(self.k0, kmax, self._pos_1[it])

            # Tabulate densities rho_0, rho_1
            for kk, knorm in enumerate(kgrid):
                for i in selection[kk]:
                    ik = self.kvector[knorm][i]
                    if ndims == 3:
                        rho_0[it][ik] = numpy.sum(expo_0[..., 0, ik[0]] * expo_0[..., 1, ik[1]] * expo_0[..., 2, ik[2]])
                    elif ndims == 2:
                        rho_0[it][ik] = numpy.sum(expo_0[..., 0, ik[0]] * expo_0[..., 1, ik[1]])
                    else:
                        # Arbitrary dimension (a bit slower)
                        tmp = expo_0[..., 0, ik[0]]
                        for idim in range(1, len(ik)):
                            tmp *= expo_0[..., idim, ik[idim]]
                        rho_0[it][ik] += numpy.sum(tmp).real                            
                        
                    # Same optimization as above: only calculate rho_1 if needed
                    if self._pos_1 is not self._pos_0:
                        if ndims == 3:
                            rho_1[it][ik] = numpy.sum(expo_1[..., 0, ik[0]] * expo_1[..., 1, ik[1]] * expo_1[..., 2, ik[2]])
                        elif ndims == 2:
                            rho_1[it][ik] = numpy.sum(expo_1[..., 0, ik[0]] * expo_1[..., 1, ik[1]])
                        else:
                            # Arbitrary dimension (a bit slower)
                            tmp = expo_1[..., 0, ik[0]]
                            for idim in range(1, len(ik)):
                                tmp *= expo_1[..., idim, ik[idim]]
                            rho_1[it][ik] += numpy.sum(tmp).real                            

            # Optimization
            if self._pos_1 is self._pos_0:
                rho_1 = rho_0

        return rho_0, rho_1

    def _compute(self):
        # Setup k vectors and tabulate densities
        kgrid, selection =  self.kgrid, self.selection
        rho_0, rho_1 = self._tabulate_rho(kgrid, selection)

        # Compute correlation function
        acf = [defaultdict(float) for _ in kgrid]
        cnt = [defaultdict(float) for _ in kgrid]
        skip = self.skip
        for kk, knorm in enumerate(progress(kgrid)):
            for j in selection[kk]:
                ik = self.kvector[knorm][j]
                for off, i in self._discrete_tgrid:
                    for i0 in range(off, len(rho_0)-i, skip):
                        # Get the actual time difference
                        # TODO: It looks like the order of i0 and ik lopps should be swapped
                        dt = self.trajectory.steps[i0+i] - self.trajectory.steps[i0]
                        acf[kk][dt] += (rho_0[i0+i][ik] * rho_1[i0][ik].conjugate()).real #/ self._pos[i0].shape[0]
                        cnt[kk][dt] += 1

        # Normalization
        times = sorted(acf[0].keys())
        self.grid[0] = kgrid
        self.grid[1] = [ti*self.trajectory.timestep for ti in times]
        # TODO: check normalization when not GC, does not give exactly the short time behavior as pp.x
        # This switch is not used
        # if self._pos_0 is self._pos_1:
        #     nav = sum([p.shape[0] for p in self._pos]) / len(self._pos)
        # else:
        #     nav_0 = sum([p.shape[0] for p in self._pos_0]) / len(self._pos_0)
        #     nav_1 = sum([p.shape[0] for p in self._pos_1]) / len(self._pos_1)
        # First normalize by cnt (time counts), then by value at t=0
        # We do not need to normalize by the average number of particles
        self.value_nonorm = [[acf[kk][ti] / (cnt[kk][ti]) for ti in times] for kk in range(len(self.grid[0]))]

        # Normalize
        if self.normalize:
            self.value = [[v / self.value_nonorm[kk][0] for v in self.value_nonorm[kk]] for kk in range(len(self.grid[0]))]
        else:
            self.value = self.value_nonorm

    def analyze(self):
        self.analysis['relaxation times tau'] = _extract_tau(self.grid[0], self.grid[1], self.value)

    def write(self):
        Correlation.write(self)
        if self._output_file != '/dev/stdout':
            with open(self._output_file + '.tau', 'w') as out:
                _write_tau(out, self.analysis)



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Fourier-space post processing code."""

import math
import logging
import random
from collections import defaultdict

import numpy

from .helpers import linear_grid
from .correlation import Correlation

__all__ = ['expo_sphere', 'expo_sphere_safe', 'FourierSpaceCorrelation']

_log = logging.getLogger(__name__)


def expo_sphere(k0, kmax, pos):

    """Returns the exponentials of the input positions for each k."""

    # Technical note: we use ellipsis, so that we can pass either a
    # single sample or multiple samples without having to add a
    # trivial extra dimension to input array
    im = numpy.complex(0.0, 1.0)
    # The integer grid must be the same as the one set in kgrid,
    # otherwise there is an offset the problem is that integer
    # negative indexing is impossible in python and rounding or
    # truncating kmax can slightly offset the grid

    # We pick up the smallest k0 to compute the integer grid
    # This leaves many unused vectors in the other directions, which
    # could be dropped using different nkmax for x, y, z
    nk_max = 1 + int(kmax / min(k0))
    # The shape of expo is nframes, N, ndim, 2*nk+1
    expo = numpy.ndarray((len(pos), ) + pos[0].shape + (2*nk_max+1, ), numpy.complex)
    expo[..., nk_max] = numpy.complex(1.0, 0.0)
    # First fill positive k
    for j in range(pos[0].shape[-1]):
        expo[..., j, nk_max+1] = numpy.exp(im * k0[j] * pos[..., j])
        expo[..., j, nk_max-1] = expo[..., j, nk_max+1].conjugate()
        for i in range(2, nk_max):
            expo[..., j, nk_max+i] = expo[..., j, nk_max+i-1] * expo[..., j, nk_max+1]
    # Then take complex conj for negative ones
    for i in range(2, nk_max+1):
        expo[..., nk_max+i] = expo[..., nk_max+i-1] * expo[..., nk_max+1]
        expo[..., nk_max-i] = expo[..., nk_max+i].conjugate()

    return expo


def expo_sphere_safe(k0, kmax, pos):
    """
    Returns the exponentials of the input positions for each k.
    It does not use ellipsis.
    """
    im = numpy.complex(0.0, 1.0)
    ndims = pos.shape[-1]
    nk_max = 1 + int(kmax / min(k0))
    expo = numpy.ndarray(pos.shape + (2*nk_max+1, ), numpy.complex)
    expo[:, :, :, nk_max] = numpy.complex(1.0, 0.0)

    for j in range(ndims):
        expo[:, :, j, nk_max+1] = numpy.exp(im*k0[j]*pos[:, :, j])
        expo[:, :, j, nk_max-1] = expo[:, :, j, nk_max+1].conjugate()
        for i in range(2, nk_max):
            expo[:, :, j, nk_max+i] = expo[:, :, j, nk_max+i-1] * expo[:, :, j, nk_max+1]

    for i in range(2, nk_max+1):
        expo[:, :, :, nk_max+i] = expo[:, :, :, nk_max+i-1] * expo[:, :, :, nk_max+1]
        expo[:, :, :, nk_max-i] = expo[:, :, :, nk_max+i].conjugate()

    return expo


def _k_norm(ik, k0, offset):
    k_shift = k0 * (numpy.array(ik) - offset)
    k_sq = numpy.dot(k_shift, k_shift)
    return math.sqrt(k_sq)

def _sphere(kmax):
    ikvec = numpy.ndarray(3, dtype=int)
    for ix in range(-kmax, kmax+1):
        for iy in range(-kmax, kmax+1):
            for iz in range(-kmax, kmax+1):
                ikvec[0] = ix
                ikvec[1] = iy
                ikvec[2] = iz
                yield ikvec
                
def _disk(kmax):
    ikvec = numpy.ndarray(2, dtype=int)
    for ix in range(-kmax, kmax+1):
        for iy in range(-kmax, kmax+1):
            ikvec[0] = ix
            ikvec[1] = iy
            yield ikvec
                        

class FourierSpaceCorrelation(Correlation):

    """
    Base class for Fourier space correlation functions.

    The correlation function is computed for each of the scalar values
    k_i of the provided `kgrid`. If the latter is `None`, the grid is
    built using `ksamples` entries linearly spaced between `kmin` and
    `kmax`.

    For each sample k_i in `kgrid`, the correlation function is
    computed over at most `nk` wave-vectors (k_x, k_y, k_z) such that
    their norm (k_x^2+k_y^2+k_z^2)^{1/2} lies within `dk` of the
    prescribed value k_i.

    See the doc of `Correlation` for information about the rest of the
    instance variables.
    """

    def __init__(self, trajectory, grid, norigins=None, nk=8, dk=0.1,
                 kmin=-1, kmax=10, ksamples=20, fix_cm=False, normalize=True):
        super(FourierSpaceCorrelation, self).__init__(trajectory,
                                                      grid, norigins=norigins, fix_cm=fix_cm)
        # Some additional variables. k0 = smallest wave vectors
        # compatible with the boundary conditions
        # TODO: document the additional data structures used to store k vectors
        # TODO: streamline k vectors data structures
        self.normalize = normalize
        self.nk = nk
        self.dk = dk
        self.kmin = kmin
        self.kmax = kmax
        self.ksamples = ksamples
        self.kgrid = []
        self.k0 = []
        self.kvector = {}
        self.selection = []
        self._kbin_max = 0

    def compute(self):
        # We subclass compute to define k grid at compute time
        # Find k-norms grid and store it a self.kgrid (the norms are sorted)
        variables = self.short_name.split('(')[1][:-1]
        variables = variables.split(',')
        if len(variables) > 1:
            self.kgrid = self.grid[variables.index('k')]
        else:
            self.kgrid = self.grid

        # Setup grid once. If cell changes we'll call it again
        self._setup()

        # Pick up a random, unique set of nk vectors out ot the avilable ones
        # without exceeding maximum number of vectors in shell nkmax
        self.kgrid, self.selection = self._decimate_k()
        # We redefine the grid because of slight differences on the
        # average k norms appear after decimation.
        self.kgrid = self._actual_k_grid()
        # We must fix the keys: just pop them to the their new positions
        # We sort both of them (better check len's)
        for k, kv in zip(sorted(self.kgrid), sorted(self.kvector)):
            self.kvector[k] = self.kvector.pop(kv)

        # Now compute
        super(FourierSpaceCorrelation, self).compute()

    def _setup(self, sample=0):
        self.k0 = 2*math.pi/self.trajectory[sample].cell.side
        # If grid is not provided, setup a linear grid from kmin,kmax,ksamples data
        # TODO: This shouldnt be allowed with fluctuating cells
        # Or we should fix the smallest k to some average of smallest k per sample
        if self.kgrid is None:
            if self.kmin > 0:
                self.kgrid = linear_grid(self.kmin, self.kmax, self.ksamples)
            else:
                self.kgrid = linear_grid(min(self.k0), self.kmax, self.ksamples)
        else:
            # Sort, since code below depends on kgrid[0] being the smallest k-value.
            self.kgrid.sort()
            # If the first wave-vector is negative we replace it by k0
            if self.kgrid[0] < 0.0:
                self.kgrid[0] = min(self.k0)

        # Setup the grid of wave-vectors
        self.kvector, self._kbin_max = self._setup_grid_sphere(len(self.kgrid) * [self.dk],
                                                               self.kgrid, self.k0)
        
    @staticmethod
    def _setup_grid_sphere(dk, kgrid, k0):
        """
        Setup wave vector grid with spherical average (no symmetry),
        picking up vectors that fit into shells of width dk centered around
        the values specified in the input list kgrid.
        Returns a dictonary of lists of wavevectors, one entry for each element in the grid.
        """
        _log.info('setting up the wave-vector grid')
        kvec = defaultdict(list)
        # With elongated box, we choose the smallest k0 component to
        # setup the integer grid. This must be consistent with
        # expo_grid() otherwise it wont find the vectors
        kmax = kgrid[-1] + dk[-1]
        kbin_max = 1 + int(kmax / min(k0))
        kmax_sq = kmax**2

        # Choose iterator of spatial grid
        ndims = len(k0)
        if ndims == 3:
            _iterator = _sphere
        elif ndims == 2:
            _iterator = _disk
        else:
            raise ValueError('unsupported dimension {}'.format(ndims))
            
        for ik in _iterator(kbin_max):
            ksq = numpy.dot(k0*ik, k0*ik)
            if ksq > kmax_sq:
                continue
            # beware: numpy.sqrt is x5 slower than math one!
            knorm = math.sqrt(ksq)
            # Look for a shell of vectors in which the vector could fit.
            # This expression is general and allows arbitrary k grids
            # However, searching for the shell like this is not fast
            # (it costs about as much as the above)
            for ki, dki in zip(kgrid, dk):
                if abs(knorm - ki) < dki:
                    kvec[ki].append(tuple(ik + kbin_max))
                    break

        if len(kvec.keys()) != len(kgrid):
            _log.warning('some entries in the kgrid had no matching k-vector')

        return kvec, kbin_max

    def _decimate_k(self):
        """
        Pick up a random, unique set of nk vectors out ot the avilable
        ones without exceeding maximum number of vectors in shell
        nkmax.

        Return: the new list of k-norms and a list of random samples
        of k-vector indices.
        """
        # Setting the seed here once so as to get the same set
        # independent of filters.
        random.seed(1)
        kgrid = sorted(self.kvector.keys())
        selection = []
        for knorm in kgrid:
            nkmax = len(self.kvector[knorm])
            selection.append(random.sample(list(range(nkmax)), min(self.nk, nkmax)))
        return kgrid, selection

    @property
    def kvectors(self):
        """
        Dictionary of actual kvectors used to compute the correlation
        function

        The keys of the dictionary are the kgrid values, i.e. the
        average k values in each wave-vector shell. The values are
        lists of wavevectors.
        """
        db = defaultdict(list)
        for ik, knorm in enumerate(self.kgrid):
            for isel in self.selection[ik]:
                center_vec = numpy.array(self.kvector[knorm][isel]) - self._kbin_max
                db[knorm].append(list(self.k0 * center_vec))
        return db
    
    def report(self, verbose=False):
        """
        Return a formatted report of the wave-vector grid used to compute
        the correlation function

        The `verbose` option turns on writing of the individuals
        wavectors (also accessible via the `kvectors` property).
        """
        txt = '# k-point, average, std, vectors in shell\n'
        db = self.kvectors
        for knorm in db:
            knorms = []
            for kvec in db[knorm]:
                k_sq = numpy.dot(kvec, kvec)
                knorms.append(math.sqrt(k_sq))
            knorms = numpy.array(knorms)
            txt += "{} {:f} {:f} {}\n".format(knorm, knorms.mean(),
                                              knorms.std(),
                                              len(db[knorm]))
        if verbose:
            txt += '\n# k-point, k-vector\n'
            for knorm in db:
                for kvec in db[knorm]:
                    # Reformat numpy array
                    as_str = str(kvec)
                    as_str = as_str.replace(',', '')
                    as_str = as_str.replace('[', '')
                    as_str = as_str.replace(']', '')
                    txt += '{} {}\n'.format(knorm, as_str)
        return txt

    def _actual_k_grid(self):
        """
        Return exactly the average wave vectors from the selected ones

        Used to redefine the k grid.
        """
        k_grid = []
        for kk, knorm in enumerate(self.kgrid):
            av = 0.0
            for i in self.selection[kk]:
                av += _k_norm(self.kvector[knorm][i], self.k0, self._kbin_max)
            k_grid.append(av / len(self.selection[kk]))
        return k_grid



#!/usr/bin/env python

import os
import glob

# We use numpy distutils to compile and wrap f90 code via f2py
import setuptools
    
with open('README.md', 'r') as fh:
    readme = fh.read()

with open('atooms/postprocessing/_version.py') as f:
    exec(f.read())

args = dict(name='atooms-pp',
            version=__version__,
            description='Post-processing tools for particle simulations',
            long_description=readme,
            long_description_content_type="text/markdown",
            author='Daniele Coslovich',
            author_email='daniele.coslovich@umontpellier.fr',
            url='http://www.coulomb.univ-montp2.fr/perso/daniele.coslovich/',
            packages=['atooms', 'atooms/postprocessing'],
            scripts=glob.glob(os.path.join('bin', '*.py')),
            install_requires=['atooms>=1.10,<3', 'numpy', 'argh', 'tqdm'],
            license='GPLv3',
            setup_requires = ['numpy'],
            classifiers=[
                'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
                'Development Status :: 5 - Production/Stable',
                'Intended Audience :: Science/Research',
                'Programming Language :: Python :: 2',
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3',
                'Programming Language :: Python :: 3.4',
                'Topic :: Scientific/Engineering :: Physics',
            ]
)

try:
    from numpy.distutils.core import setup, Extension
    
    args["ext_modules"] = [Extension('atooms.postprocessing.realspace_wrap',
                                     sources=['atooms/postprocessing/realspace.f90'],
                                     extra_f90_compile_args=[]),
                           Extension('atooms.postprocessing.fourierspace_wrap', 
                                     sources=['atooms/postprocessing/fourierspace.f90'],
                                     extra_f90_compile_args=[])]

except (ModuleNotFoundError, ImportError):
    from distutils.core import setup

setup(**args)



#!/usr/bin/env python

import sys
import numpy
import atooms.postprocessing
from atooms.postprocessing.helpers import linear_grid, logx_grid

def main(fname):
    x = numpy.loadtxt(sys.stdin if fname=='-' else fname, unpack=True)
    t_grid = list(range(0, len(x), 10))
    t = list(range(len(x)))
    x, y, z = postprocessing.correlation.acf(t_grid, 1, t, x)
    for xi, yi in zip(x, y):
        print(xi, yi)

if __name__ == '__main__':

    import sys
    from optparse import OptionParser

    parser = OptionParser()
    parser.add_option('--vacf-samples', action='store',   dest='vacf_samples', default=60,   type='int')
    parser.add_option('--vacf-tmax', action='store',      dest='vacf_tmax', default=1.0, type='float')
    (opts, args) = parser.parse_args()
    
    main('-')
        



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Velocity autocorrelation function."""

import numpy

from .correlation import Correlation, gcf_offset
from .helpers import setup_t_grid

__all__ = ['VelocityAutocorrelation']


class VelocityAutocorrelation(Correlation):

    """Velocity autocorrelation function."""

    symbol = 'vacf'
    short_name = 'Z(t)'
    long_name = 'velocity autocorrelation'
    phasespace = ['vel']

    def __init__(self, trajectory, tgrid=None, norigins=None):
        Correlation.__init__(self, trajectory, tgrid, norigins=norigins)
        if self.grid is None:
            self.grid = linear_grid(0.0, trajectory.total_time * 0.10, tsamples)
        self._discrete_tgrid = setup_t_grid(self.trajectory, tgrid, offset=norigins != '1')

    def _compute(self):
        def f(x, y):
            return numpy.sum(x * y) / float(x.shape[0])
        self.grid, self.value = gcf_offset(f, self._discrete_tgrid,
                                           self.trajectory.block_size,
                                           self.trajectory.steps,
                                           self._vel)
        self.grid = [ti * self.trajectory.timestep for ti in self.grid]



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Radial distribution function."""

import math
import logging

import numpy

from .helpers import linear_grid
from .correlation import Correlation
from .progress import progress

__all__ = ['RadialDistributionFunction',
           'RadialDistributionFunctionLegacy',
           'RadialDistributionFunctionFast']

_log = logging.getLogger(__name__)


def gr_kernel(x, y, L, *args):
    # Precalculating 1/L does not improve timings
    # r is an array of array distances
    r = x-y
    r = r - numpy.rint(r/L) * L
    return numpy.sqrt(numpy.sum(r**2, axis=1))


def gr_kernel_square(x, y, L, *args):
    """Return square distances."""
    # r is an array of array distances
    r = x-y
    r = r - numpy.rint(r/L) * L
    return numpy.sum(r**2, axis=1)

def pairs_newton_hist(f, x, y, L, bins):
    """
    Apply function f to all pairs in x[i] and y[j] and update the
    |hist| histogram using the |bins| bin edges.
    """
    hist, bins = numpy.histogram([], bins)
    # Do the calculation in batches to optimize
    bl = max(1, int(1e5 / len(y)))
    for ib in range(0, len(y)-1, bl):
        fxy = []
        # batch must never exceed len(y)-1
        for i in range(ib, min(ib+bl, len(y)-1)):
            for value in f(x[i+1:], y[i], L):
                fxy.append(value)
        hist_tmp, bins = numpy.histogram(fxy, bins)
        hist += hist_tmp
    return hist


def pairs_hist(f, x, y, L, bins):
    """
    Apply function f to all pairs in x[i] and y[j] and update the
    |hist| histogram using the |bins| bin edges.
    """
    hist, bins = numpy.histogram([], bins)
    for i in range(len(y)):
        fxy = f(x[:], y[i], L)
        hist_tmp, bins = numpy.histogram(fxy, bins)
        hist += hist_tmp
    return hist


class RadialDistributionFunctionLegacy(Correlation):
    """
    Radial distribution function.

    The correlation function g(r) is computed over a grid of distances
    `rgrid`. If the latter is `None`, the grid is linear from 0 to L/2
    with a spacing of `dr`. Here, L is the side of the simulation cell
    along the x axis at the first step.

    Additional parameters:
    ----------------------

    - norigins: controls the number of trajectory frames to compute
      the time average
    """

    nbodies = 2
    symbol = 'gr'
    short_name = 'g(r)'
    long_name = 'radial distribution function'
    phasespace = 'pos'

    def __init__(self, trajectory, rgrid=None, norigins=None, dr=0.04, ndim=-1, rmax=-1.0):
        Correlation.__init__(self, trajectory, rgrid, norigins=norigins)
        self.dr = dr
        self.rmax = rmax
        """
        Limit distance binning up to `rmax`. It may enable linked cells if
        this is advantageous.
        """

    def _setup_grid(self, system):
        """
        Deferred grid setup
        """
        if self.grid is None:
            # Get side of cell. If the system does not have a cell,
            # wrap it in an infinite cell and estimate a reasonable grid.
            ndims = system.number_of_dimensions
            if system.cell is not None:
                # Redefine grid to extend up to L
                # The grid will be cropped later to L/2
                # This retains the original dr
                self.grid = linear_grid(self.dr/2, min(system.cell.side), self.dr)
            else:
                # If there is no cell, then rmax must have been given
                # This retains the original dr
                if self.rmax > 0:
                    self.grid = linear_grid(self.dr/2, self.rmax, self.dr)
                else:
                    self.grid = linear_grid(self.dr/2, self.dr*1000, self.dr)

        # Internal max distance (/= user provided rmax)
        self._rmax = self.grid[-1]
        
    def _compute(self):
        ncfg = len(self.trajectory)
        system = self.trajectory.read(0)
        ndims = system.number_of_dimensions
        self._setup_grid(system)
        
        # Reconstruct bounds of grid for numpy histogram
        grid = []
        for i in range(len(self.grid)):
            grid.append(self.grid[i] - (self.grid[1] - self.grid[0]) / 2)
        grid.append(self.grid[-1] + (self.grid[1] - self.grid[0]) / 2)
        self.grid = grid
        gr, bins = numpy.histogram([], bins=self.grid)

        # Assume grandcanonical trajectory for generality.
        # Note that testing if the trajectory is grandcanonical or
        # semigrandcanonical is useless when applying filters.
        # N_0, N_1 = len(self._pos_0[0]), len(self._pos_1[0])
        N_0 = numpy.average([len(x) for x in self._pos_0])
        N_1 = numpy.average([len(x) for x in self._pos_1])

        gr_all = []
        _, r = numpy.histogram([], bins=self.grid)
        origins = range(0, ncfg, self.skip)
        for i in progress(origins):
            system = self.trajectory.read(i)
            side = system.cell.side
            if len(self._pos_0[i]) == 0 or len(self._pos_1[i]) == 0:
                continue
            if self._pos_0 is self._pos_1:
                gr = pairs_newton_hist(gr_kernel, self._pos_0[i], self._pos_1[i], side, r)
            else:
                gr = pairs_hist(gr_kernel, self._pos_0[i], self._pos_1[i], side, r)
            gr_all.append(gr)

        # Normalization
        volume = system.cell.volume
        if ndims == 2:
            vol = math.pi * (r[1:]**2 - r[:-1]**2)
        elif ndims == 3:
            vol = 4 * math.pi / 3.0 * (r[1:]**3 - r[:-1]**3)
        else:
            from math import gamma
            n2 = int(float(ndims) / 2)
            vol = math.pi**n2 * (r[1:]**ndims-r[:-1]**ndims) / gamma(n2+1)
        rho = N_1 / volume
        if self._pos_0 is self._pos_1:
            norm = rho * vol * N_0 * 0.5  # use Newton III
        else:
            norm = rho * vol * N_0
        gr = numpy.average(gr_all, axis=0)
        self.grid = (r[:-1] + r[1:]) / 2.0
        self.value = gr / norm
        # Restrict distances to L/2 (in last frame) or rmax 
        if self.rmax > 0:
            where = self.grid <= self.rmax
        else:
            side = system.cell.side
            where = self.grid <= min(side) / 2
        self.grid = self.grid[where]
        self.value = self.value[where]


class RadialDistributionFunctionFast(RadialDistributionFunctionLegacy):
    """
    Radial distribution function using f90 kernel.

    The correlation function g(r) is computed over a grid of distances
    `rgrid`. If the latter is `None`, the grid is linear from 0 to L/2
    with a spacing of `dr`. Here, L is the side of the simulation cell
    along the x axis at the first step.

    Additional parameters:
    ----------------------

    - norigins: controls the number of trajectory frames to compute
      the time average
    """

    def _compute(self):
        from atooms.postprocessing.realspace_wrap import compute
        from atooms.postprocessing.linkedcells import LinkedCells

        # Assume grandcanonical trajectory for generality.
        # Note that testing if the trajectory is grandcanonical or
        # semigrandcanonical is useless when applying filters.
        N_0, N_1 = [], []
        gr_all = []

        # Grid setup.
        # Get side of cell. If the system does not have a cell,
        # wrap it in an infinite cell and estimate a reasonable grid.
        system = self.trajectory.read(0)
        ndims = system.number_of_dimensions
        self._setup_grid(system)
        dr = self.grid[1] - self.grid[0]
        gr = numpy.zeros(len(self.grid), dtype=int)
        bins = numpy.array(self.grid)
        
        # Use linked cells only if it is advantageous
        # - more than 3 cells along each side
        # - memory footprint is < ~1Gb
        # These tests are done of the first framce
        # TODO: if memory footprint is surpassed skip particles
        linkedcells = None
        if self.rmax > 0.0 and system.cell is not None:
            n_0 = len(self._pos_0[0])
            n_1 = len(self._pos_1[0])
            rho = n_1 / system.cell.volume
            nmax = self.rmax**ndims * rho
            if int(min(system.cell.side / self.rmax)) > 3 and nmax * n_0 < 1e6:
                _log.info('using linked cells')
                linkedcells = LinkedCells(rcut=self.rmax)
            else:
                _log.info('not using linked cells')
                linkedcells = None

        # Main loop for average
        origins = range(0, len(self.trajectory), self.skip)
        for i in progress(origins):
            # Skip if there are no particles
            if len(self._pos_0[i]) == 0 or len(self._pos_1[i]) == 0:
                continue

            # Switch between self and distinct calculation
            distinct = self._pos_0 is not self._pos_1

            # Store side
            system = self.trajectory.read(i)
            if system.cell is not None:
                side = system.cell.side
            else:
                import sys
                side = numpy.ndarray(ndims, dtype=float)
                side[:] = sys.float_info.max

            # When using linked cells, we precalculate the neighbors
            if linkedcells:
                if self._pos_0 is self._pos_1:
                    neighbors, number_of_neighbors = linkedcells.compute(side, self._pos_0[i], as_array=True)
                else:
                    neighbors, number_of_neighbors = linkedcells.compute(side, self._pos_0[i], self._pos_1[i], as_array=True)

            # Transpose the arrays for fortran
            # TODO: can we avoid this?
            if distinct:
                pos_0 = self._pos_0[i].transpose()
                pos_1 = self._pos_1[i].transpose()
            else:
                pos_0 = self._pos_0[i].transpose()
                pos_1 = pos_0

            # With a non-periodic cell, which just bounds the physical
            # domain, we must crop particles close to the surface
            # self_term = 0
            if hasattr(system.cell, 'periodic') and not numpy.any(system.cell.periodic):
                #mask = numpy.ndarray(pos_0.shape[1], dtype=numpy.bool)
                # TODO: bool does not work
                mask = compute.on_surface(pos_0, side, self.rmax)
                mask = mask == 1
                pos_0 = pos_0[:, mask]
                # Force distinct calculation
                distinct = True

            # Store number of particles for normalization
            N_0.append(pos_0.shape[1])
            N_1.append(pos_1.shape[1])

            # Compute g(r)
            if not distinct:
                if linkedcells is None:
                    compute.gr_self(pos_0, side, dr, gr, bins)
                else:
                    compute.gr_neighbors_self('C', pos_0, neighbors, number_of_neighbors, side, dr, gr, bins)
            else:
                if linkedcells is None:
                    compute.gr_distinct(pos_0, pos_1, side, dr, gr, bins)
                else:
                    compute.gr_neighbors_distinct('C', pos_0, pos_1, neighbors, number_of_neighbors, side, dr, gr, bins)

            # Damned copies in python
            gr_all.append(gr.copy())

        # Normalization
        # Array r is used to compute shells, we thus use the bin boundaries
        r = bins - (bins[1] - bins[0]) / 2
        N_0 = numpy.average(N_0)
        N_1 = numpy.average(N_1)
        if system.cell is not None:
            volume = system.cell.volume
        else:
            # TODO: this only works with recent atooms
            volume = len(system.particle) / system.density
        if ndims == 2:
            vol = math.pi * (r[1:]**2 - r[:-1]**2)
        elif ndims == 3:
            vol = 4 * math.pi / 3.0 * (r[1:]**3 - r[:-1]**3)
        else:
            from math import gamma
            n2 = int(float(ndims) / 2)
            vol = math.pi**n2 * (r[1:]**ndims-r[:-1]**ndims) / gamma(n2+1)

        rho = N_1 / volume
        norm = rho * vol * N_0
        if not distinct:
            # We used Newton III
            norm /= 2
        gr = numpy.average(gr_all, axis=0)

        # We crop the last entry
        self.grid = bins[:-1]
        self.value = gr[:-1] / norm

        # Restrict distances to L/2 (in last frame) or rmax 
        if self.rmax > 0:
            where = self.grid <= self.rmax
        else:
            side = system.cell.side
            where = self.grid <= min(side) / 2
        self.grid = self.grid[where]
        self.value = self.value[where]

# Defaults to fast
try:
    import atooms.postprocessing.realspace_wrap
    RadialDistributionFunction = RadialDistributionFunctionFast
except ImportError:
    RadialDistributionFunction = RadialDistributionFunctionLegacy



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Free volume spectral density."""

import numpy
from atooms.trajectory import Trajectory
from atooms.trajectory.utils import is_cell_variable

from .fourierspace import FourierSpaceCorrelation, expo_sphere

__all__ = ['SpectralDensity']


class SpectralDensity(FourierSpaceCorrelation):
    """
    Free volume spectral density.

    From Zachary, Jiao, Torquato PRL 106, 178001 (2011).

    See the documentation of the `FourierSpaceCorrelation` base class
    for information on the instance variables.
    """

    symbol = 'ik'
    short_name = 'I(k)'
    long_name = 'spectral density'
    phasespace = 'pos'

    def __init__(self, trajectory, trajectory_radius, kgrid=None,
                 norigins=-1, nk=20, dk=0.1, kmin=-1.0, kmax=15.0,
                 ksamples=30):
        FourierSpaceCorrelation.__init__(self, trajectory, kgrid, norigins, nk,
                                         dk, kmin, kmax, ksamples)
        self._is_cell_variable = None
        # TODO: check step consistency 06.09.2017
        with Trajectory(trajectory_radius) as th:
            self._radius = [s.dump('particle.radius') for s in th]

    def _compute(self):
        nsteps = len(self._pos)
        # Setup k vectors and tabulate rho
        kgrid, selection = self.kgrid, self.selection
        kmax = max(self.kvector.keys()) + self.dk
        cnt = [0 for k in kgrid]
        # Note: actually rho_av is not calculated because it is negligible
        rho_av = [complex(0., 0.) for k in kgrid]
        rho2_av = [complex(0., 0.) for k in kgrid]
        cell_variable = is_cell_variable(self.trajectory)
        for i in range(0, nsteps, self.skip):
            # If cell changes we have to update
            if cell_variable:
                self._setup(i)
                kgrid, selection = self._decimate_k()
                kmax = max(self.kvector.keys()) + self.dk

            expo = expo_sphere(self.k0, kmax, self._pos[i])
            for kk, knorm in enumerate(kgrid):
                for k in selection[kk]:
                    ik = self.kvector[knorm][k]
                    Ri = self._radius[i]
                    mk = 4 * numpy.pi / knorm**3 * (numpy.sin(knorm*Ri) - (knorm*Ri) * numpy.cos(knorm*Ri))
                    rho = numpy.sum(mk * expo[..., 0, ik[0]] * expo[..., 1, ik[1]] * expo[..., 2, ik[2]])
                    rho2_av[kk] += (rho * rho.conjugate())
                    cnt[kk] += 1

        # Normalization.
        volume = numpy.average([s.cell.volume for s in self.trajectory])
        self.grid = kgrid
        self.value = [(rho2_av[kk] / cnt[kk] - rho_av[kk]*rho_av[kk].conjugate() / cnt[kk]**2).real / volume
                      for kk in range(len(self.grid))]
        self.value_nonorm = [rho2_av[kk].real / cnt[kk]
                             for kk in range(len(self.grid))]



# This file is part of atooms
# Copyright 2010-2018, Daniele Coslovich

"""Base correlation function."""

import os
import logging
from collections import defaultdict

import numpy
from atooms.trajectory import Trajectory
from atooms.trajectory.decorators import Unfolded, fold
from atooms.trajectory.decorators import change_species
from atooms.system.particle import distinct_species
from atooms.core.utils import Timer
try:
    from medepy.metadata import dump as _dump
except ImportError:
    from .helpers import _dump

from . import core
from .helpers import adjust_skip
from .progress import progress

__all__ = ['acf', 'gcf', 'gcf_offset', 'Correlation']

_log = logging.getLogger(__name__)
#pp_output_path = '{trajectory.filename}.pp.{symbol}.{tag}'
#pp_trajectory_format = None


def acf(grid, skip, t, x):
    """
    Auto correlation function.

    Calculate the correlation between time t(i0) and t(i0+i)
    for all possible pairs (i0,i) provided by grid.
    """
    cf = defaultdict(float)
    cnt = defaultdict(int)
    xave = numpy.average(x)
    for i in grid:
        for i0 in range(0, len(x)-i, skip):
            # Get the actual time difference
            dt = t[i0+i] - t[i0]
            cf[dt] += (x[i0+i]-xave) * (x[i0]-xave)
            cnt[dt] += 1

    # Return the ACF with the time differences sorted
    dt = sorted(cf.keys())
    return dt, [cf[ti] / cnt[ti] for ti in dt], cnt


def gcf(f, grid, skip, t, x):
    """
    Generalized auto-correlation function <f(x[i], x[i0])>

    Pass a function `f` to apply to the data at each frame.

    Exemple: mean square displacement.
    """
    # Calculate the correlation between time t(i0) and t(i0+i)
    # for all possible pairs (i0,i) provided by grid
    cf = defaultdict(float)
    cnt = defaultdict(int)
    for i in grid:
        # Note: len(x) gives x.shape[0]
        for i0 in progress(range(0, len(x)-i-1, skip)):
            # Get the actual time difference
            dt = t[i0+i] - t[i0]
            cf[dt] += f(x[i0+i], x[i0])
            cnt[dt] += 1

    # Return the ACF with the time differences sorted
    dt = sorted(cf.keys())
    return dt, [cf[ti] / cnt[ti] for ti in dt], [cnt[ti] for ti in dt]


def gcf_offset(f, grid, skip, t, x):
    """
    Generalized auto-correlation function <f(x[i], x[i0])> using a grid with offsets

    Pass a function `f` to apply to the data `x` at each frame.

    Exemple: mean square displacement.
    """
    # Calculate the correlation between time t(i0) and t(i0+i)
    # for all possible pairs (i0,i) provided by grid
    cf = defaultdict(float)
    cnt = defaultdict(int)
    # Standard calculation
    for off, i in progress(grid, total=len(grid)):
        for i0 in range(off, len(x)-i, skip):
            # Get the actual time difference
            dt = t[i0+i] - t[i0]
            cf[dt] += f(x[i0+i], x[i0])
            cnt[dt] += 1

    # Return the ACF with the time differences sorted
    dt = sorted(cf.keys())
    return dt, [cf[ti] / cnt[ti] for ti in dt]


def _subtract_mean(weight):
    mean = 0
    for current_field in weight:
        mean += current_field.mean()
    mean /= len(weight)
    for current_field in weight:
        current_field -= mean
    return weight


class Correlation(object):
    """
    Base class for correlation functions.

    The correlation function is calculated for the trajectory `trj`. This can be:

    - an object implementing the atooms `Trajectory` interface
    - the path to a trajectory file in a format recognized by atooms

    A correlation function A(x) is defined over a grid of real
    entries {x_i} given by the list `grid`. To each entry of the grid,
    the correlation function has a corresponding value A_i=A(x_i). The
    latter values are stored in the `value` list.

    Correlation functions that depend on several variables, A(x,y,...)
    must provide a list of grids, one for each variable. The order is
    one specified by the `symbol` class variable, see below.

    The correlation function A is calculated as a statistical average
    over several time origins in the trajectory `trj`. The `norigins`
    variable can be used to tune the number of time origins used to
    carry out the average. `norigins` can take the following values:

    - `norigins=-1`: all origins are used
    - an integer >= 1: use only `n_origins` time origins
    - a float in the interval (0,1): use only a fraction `norigins` of frames as time origins
    - `None`: a heuristics is used to keep the product of steps times particles constant

    Subclasses must provide a symbolic expression of the correlation
    function through the `symbol` class variable. The following
    convention is used: if a correlation function A depends on
    variables x and y, then `symbol = 'A(x,y)'`.

    The `phasespace` variable allow subclasses to access a list of 2d
    numpy arrays with particles coordinates via the following private
    variables:

    - if `nbodies` is 1: `self._pos` for positions, `self._vel` for
    velocities, `self._unf_pos` for PBC-unfolded positions

    - if `nbodies` is 2, an additional suffix that take values 0 or 1
    is added to distinguish the two sets of particles,
    e.g. `self._pos_0` and `self._pos_1`
    """

    nbodies = 1
    """
    Class variable that controls the number of bodies, i.e. particles,
    associated to the correlation function. This variable controls the
    internal arrays used to compute the correlation, see `phasespace`.
    """
    static = False
    """
    Turn this to `True` is the correlation function is static,
    i.e. not time-dependent. This may enable some optimizations.
    """
    symbol = ''
    """Example: fskt"""
    short_name = ''
    """Example: F_s(k,t)"""
    long_name = ''
    """Example: Self intermediate scattering function"""
    phasespace = ['pos', 'pos-unf', 'vel']
    """
    List of strings or string among ['pos', 'pos-unf', 'vel']. It
    indicates which variables should be read from the trajectory file.
    They will be available as self._pos, self._pos_unf, self._vel.
    """

    def __init__(self, trj, grid, output_path=None, norigins=None, fix_cm=False):
        # Accept a trajectory-like instance or a path to a trajectory
        if isinstance(trj, str):
            self.trajectory = Trajectory(trj, mode='r', fmt=core.pp_trajectory_format)
        else:
            self.trajectory = trj
        self._fix_cm = fix_cm
        self._unfolded = None
        self.grid = grid
        self.value = []
        self.analysis = {}
        self.comments = None  # can be modified by user at run time
        self.tag = ''
        self.tag_description = 'the whole system'
        self.output_path = output_path if output_path is not None else core.pp_output_path
        self.skip = adjust_skip(self.trajectory, norigins)

        # Callbacks
        self._cbk = []
        self._cbk_args = []
        self._cbk_kwargs = []

        # Lists for one body correlations
        self._pos = []
        self._vel = []
        self._ids = []
        self._pos_unf = []

        # Lists for two-body correlations
        self._pos_0, self._pos_1 = [], []
        self._vel_0, self._vel_1 = [], []
        self._ids_0, self._ids_1 = [], []
        self._pos_unf_0, self._pos_unf_1 = [], []

        # Weights
        self._weight = None
        self._weight_0, self._weight_1 = None, None
        self._weight_field = None
        self._weight_fluctuations = False

    def __str__(self):
        return '{} at <{}>'.format(self.long_name, id(self))

    def add_weight(self, trajectory=None, field=None, fluctuations=False):
        """
        Add weight from the given `field` in `trajectory`

        If `trajectory` is `None`, `self.trajectory` will be used and
        `field` must be a particle property.

        If `field` is `None`, `trajectory` is assumed to be a path an
        xyz trajectory and we use the data in last column as a weight.

        If both `field` and `trajectory` are `None` the function
        returns immediately and the weight is not set.

        The optional `fluctuations` option subtracts the mean,
        calculated from the ensemble average, from the weight.
        """
        if trajectory is None and field is None:
            return

        self._weight = []
        self._weight_fluctuations = fluctuations

        # Guessing the field from the last column of an xyz file is
        # not supported anymore
        if field is None:
            raise ValueError('provide field to use as weight')
        else:
            self._weight_field = field

        # By default we use the same trajectory as for the phasespace
        if trajectory is None:
            self._weight_trajectory = self.trajectory
        else:
            self._weight_trajectory = trajectory
            # Copy over the field
            from .helpers import copy_field
            self.trajectory.add_callback(copy_field, self._weight_field, self._weight_trajectory)

        # Make sure the steps are consistent
        if self._weight_trajectory.steps != self.trajectory.steps:
            raise ValueError('inconsistency between weight trajectory and trajectory')

        # Modify tag
        fluct = 'fluctuations' if self._weight_fluctuations else ''
        self.tag_description += ' with {} {} field'.format(self._weight_field.replace('_', ' '), fluct)
        self.tag_description = self.tag_description.replace('  ', ' ')
        self.tag += '.{}_{}'.format(self._weight_field, fluct)
        self.tag.strip('_.')

    def add_filter(self, cbk, *args, **kwargs):
        """Add filter callback `cbk` along with positional and keyword arguments"""
        if len(self._cbk) > self.nbodies:
            raise ValueError('number of filters cannot exceed n. of bodies')
        self._cbk.append(cbk)
        self._cbk_args.append(args)
        self._cbk_kwargs.append(kwargs)

    def need_update(self):
        """Check if the trajectory file is newer than the output file"""
        need = True
        if os.path.exists(self._output_file) and self._output_file != '/dev/stdout':
            if os.path.getmtime(self.trajectory.filename) < \
               os.path.getmtime(self._output_file):
                need = False
        return need

    def _setup_arrays(self):
        """
        Dump positions and/or velocities at different time frames as a
        list of numpy array.
        """
        # TODO: what happens if we call compute twice?? Shouldnt we reset the arrays?
        # Ensure phasespace is a list.
        # It may not be a class variable anymore after this
        if not isinstance(self.phasespace, list) and \
           not isinstance(self.phasespace, tuple):
            self.phasespace = [self.phasespace]

        # Setup arrays
        if self.nbodies == 1:
            self._setup_arrays_onebody()
            self._setup_weight_onebody()
        elif self.nbodies == 2:
            self._setup_arrays_twobody()
            self._setup_weight_twobody()

    def _setup_arrays_onebody(self):
        """
        Setup list of numpy arrays for one-body correlations.

        We also take care of dumping the weight if needed, see
        `add_weight()`.
        """
        # Local shortcut
        th = self.trajectory

        # We unfold the trajectory if phasespace requests it and
        # unfolded positions are not available in the trajectory, or
        # if we request folded positions with fixed center of mass
        if ('pos-unf' in self.phasespace and not hasattr(th[0].particle[0], 'position_unfolded')) or \
           ('pos' in self.phasespace and self._fix_cm):
            # We unfold the positions, caching the trajectory
            if self._unfolded is None:
                self._unfolded = Unfolded(self.trajectory, fixed_cm=self._fix_cm)
                # We must fold positions back in this case
                if 'pos' in self.phasespace:
                    self._unfolded.add_callback(fold)
            # Change the local shortcut to the unfolded trajectory
            th = self._unfolded

        # Read everything except unfolded trajectories
        if 'pos' in self.phasespace or 'vel' in self.phasespace or 'ids' in self.phasespace:
            ids = distinct_species(th[0].particle)
            for s in progress(th):
                # Apply filter if there is one
                if len(self._cbk) > 0:
                    s = self._cbk[0](s, *self._cbk_args[0], **self._cbk_kwargs[0])
                if 'pos' in self.phasespace:
                    self._pos.append(s.dump('pos'))
                if 'vel' in self.phasespace:
                    self._vel.append(s.dump('vel'))
                if 'ids' in self.phasespace:
                    _ids = s.dump('species')
                    _ids = numpy.array([ids.index(_) for _ in _ids], dtype=numpy.int32)
                    self._ids.append(_ids)

        # Read unfolded positions if requested
        if 'pos-unf' in self.phasespace:
            if hasattr(th[0].particle[0], 'position_unfolded'):
                # Unfolded positions are present in the trajectory
                for s in progress(th):
                    # Apply filter if there is one
                    if len(self._cbk) > 0:
                        s = self._cbk[0](s, *self._cbk_args[0], **self._cbk_kwargs[0])
                    # Fixing the CM must be done explicitly using
                    # particle.position_unfolded because atooms.system
                    # methods work with particle.position
                    if self._fix_cm:
                        # Compute CM using unfolded positions, which is safe
                        cm = numpy.zeros_like(s.particle[0].position_unfolded)
                        mtot = 0.0
                        for p in s.particle:
                            cm += p.position_unfolded * p.mass
                            mtot += p.mass
                        cm /= mtot
                        # Subtract it
                        for p in s.particle:
                            p.position_unfolded -= cm
                    self._pos_unf.append(s.dump('particle.position_unfolded'))
            else:
                for s in progress(th):
                    # Apply filter if there is one
                    if len(self._cbk) > 0:
                        s = self._cbk[0](s, *self._cbk_args[0], **self._cbk_kwargs[0])
                    self._pos_unf.append(s.dump('pos'))

    def _setup_weight_onebody(self):
        """
        Setup list of numpy arrays for the weight, see `add_weight()`
        """
        if self._weight is None:
            return

        # Dump arrays of weights
        for s in progress(self.trajectory):
            # Apply filter if there is one
            # TODO: fix when weight trajectory does not contain actual particle info
            # It should be possible to link the weight trajectory to the trajectory
            # and return the trajectory particles with the weight
            if len(self._cbk) > 0:
                s = self._cbk[0](s, *self._cbk_args[0], **self._cbk_kwargs[0])
            current_weight = s.dump('particle.%s' % self._weight_field)
            self._weight.append(current_weight)

        # Subtract global mean
        if self._weight_fluctuations:
            _subtract_mean(self._weight)

    def _setup_weight_twobody(self):
        """
        Setup list of numpy arrays for the weight, see `add_weight()`
        """
        if self._weight is None:
            return

        self._weight = []
        self._weight_0 = []
        self._weight_1 = []

        # TODO: add checks on number of filters
        if len(self._cbk) <= 1:
            self._setup_weight_onebody()
            self._weight_0 = self._weight
            self._weight_1 = self._weight
            return

        # Dump arrays of weights
        for s in progress(self.trajectory):
            # Apply filters
            if len(self._cbk) == 2:
                s0 = self._cbk[0](s, *self._cbk_args[0], **self._cbk_kwargs[0])
                s1 = self._cbk[1](s, *self._cbk_args[1], **self._cbk_kwargs[1])
            self._weight_0.append(s0.dump('particle.%s' % self._weight_field))
            self._weight_1.append(s1.dump('particle.%s' % self._weight_field))

        # Subtract global mean
        if self._weight_fluctuations:
            _subtract_mean(self._weight_0)
            _subtract_mean(self._weight_1)

    def _setup_arrays_twobody(self):
        """Setup list of numpy arrays for two-body correlations."""
        if len(self._cbk) <= 1:
            self._setup_arrays_onebody()
            self._pos_0 = self._pos
            self._pos_1 = self._pos
            self._vel_0 = self._vel
            self._vel_1 = self._vel
            self._ids_0 = self._ids
            self._ids_1 = self._ids
            return

        if 'pos' in self.phasespace or 'vel' in self.phasespace or 'ids' in self.phasespace:
            ids = distinct_species(self.trajectory[0].particle)
            for s in progress(self.trajectory):
                s0 = self._cbk[0](s, *self._cbk_args[0], **self._cbk_kwargs[0])
                s1 = self._cbk[1](s, *self._cbk_args[1], **self._cbk_kwargs[1])
                if 'pos' in self.phasespace:
                    self._pos_0.append(s0.dump('pos'))
                    self._pos_1.append(s1.dump('pos'))
                if 'vel' in self.phasespace:
                    self._vel_0.append(s0.dump('vel'))
                    self._vel_1.append(s1.dump('vel'))
                if 'ids' in self.phasespace:
                    _ids_0 = s0.dump('species')
                    _ids_1 = s1.dump('species')
                    _ids_0 = numpy.array([ids.index(_) for _ in _ids_0], dtype=numpy.int32)
                    _ids_1 = numpy.array([ids.index(_) for _ in _ids_1], dtype=numpy.int32)
                    self._ids_0.append(_ids_0)
                    self._ids_1.append(_ids_1)

        # Dump unfolded positions if requested
        if 'pos-unf' in self.phasespace:
            for s in progress(Unfolded(self.trajectory)):
                s0 = self._cbk[0](s, *self._cbk_args[0], **self._cbk_kwargs[0])
                s1 = self._cbk[1](s, *self._cbk_args[1], **self._cbk_kwargs[1])
                self._pos_unf_0.append(s0.dump('pos'))
                self._pos_unf_1.append(s1.dump('pos'))

    def compute(self):
        """
        Compute the correlation function.

        It wraps the _compute() method implemented by subclasses.
        This method sets the `self.grid` and `self.value` variables,
        which are also returned.
        """
        _log.info('setup arrays for %s', self.tag_description)
        t = [Timer(), Timer()]
        t[0].start()
        self._setup_arrays()
        t[0].stop()

        _log.info('computing %s for %s', self.long_name, self.tag_description)
        _log.info('using %s time origins out of %s',
                  len(range(0, len(self.trajectory), self.skip)),
                  len(self.trajectory))
        t[1].start()
        self._compute()
        t[1].stop()

        _log.info('output file %s', self._output_file)
        _log.info('done %s for %s in %.1f sec [setup:%.0f%%, compute: %.0f%%]',
                  self.long_name,
                  self.tag_description, t[0].wall_time + t[1].wall_time,
                  t[0].wall_time / (t[0].wall_time + t[1].wall_time) * 100,
                  t[1].wall_time / (t[0].wall_time + t[1].wall_time) * 100)
        _log.info('')

        return self.grid, self.value

    def _compute(self):
        """Subclasses must implement this"""
        pass

    def analyze(self):
        """
        Subclasses may implement this and store the results in the
        self.analysis dictonary
        """
        pass

    @property
    def _output_file(self):
        """Returns path of output file"""
        # Interpolate the output path string
        if self.output_path is None:
            filename = None
        else:
            filename = self.output_path.format(symbol=self.symbol,
                                               short_name=self.short_name,
                                               long_name=self.long_name.replace(' ', '_'),
                                               tag=self.tag,
                                               tag_description=self.tag_description.replace(' ', '_'),
                                               trajectory=self.trajectory)
            # Strip unpleasant punctuation from basename path
            for punct in ['.', '_', '-']:
                subpaths = filename.split('/')
                subpaths[-1] = subpaths[-1].replace(punct * 2, punct)
                subpaths[-1] = subpaths[-1].strip(punct)
                filename = '/'.join(subpaths)
        return filename

    def read(self):
        """Read correlation function from existing file"""
        with open(self._output_file, 'r') as inp:
            x = numpy.loadtxt(inp, unpack=True)
            if len(x) == 3:
                _log.warn("cannot read 3-columns files yet in %s", self._output_file)
            elif len(x) == 2:
                self.grid, self.value = x
            else:
                self.grid, self.value = x[0: 2]
                _log.warn("Ignoring some columns in %s", self._output_file)

    @property
    def grid_name(self):
        """
        Return the name of the grid variables

        Example:
        -------
        If `self.name` is `F_s(k,t)`, the function returns `['k', 't']`
        """
        variables = self.short_name.split('(')[1][:-1]
        return variables.split(',')

    def write(self):
        """
        Write the correlation function and the analysis data to file

        The `output_path` instance variable is used to define the
        output files by interpolating the following variables:

        - symbol
        - short_name
        - long_name
        - tag
        - tag_description
        - trajectory

        The default is defined by core.pp_output_path, which currently
        looks like '{trajectory.filename}.pp.{symbol}.{tag}'
        """
        def is_iterable(maybe_iterable):
            try:
                iter(maybe_iterable)
            except TypeError:
                return False
            else:
                return True

        # Pack grid and value into arrays to dump
        if is_iterable(self.grid[0]) and len(self.grid) == 2:
            x = numpy.array(self.grid[0]).repeat(len(self.value[0]))
            y = numpy.array(self.grid[1] * len(self.grid[0]))
            z = numpy.array(self.value).flatten()
            dump = numpy.transpose(numpy.array([x, y, z]))
        else:
            dump = numpy.transpose(numpy.array([self.grid, self.value]))

        # Comment line
        # Extract variables from parenthesis in symbol
        variables = self.short_name.split('(')[1][:-1]
        variables = variables.split(',')
        columns = variables + [self.short_name]  #[self.symbol]
        if len(self.tag_description) > 0:
            conj = 'of'
        else:
            conj = ''
        comments = _dump(title='%s %s %s %s' % (self.long_name, self.short_name, conj, self.tag_description),
                         columns=columns,
                         command='atooms-pp', version=core.__version__,
                         description=None, note=None,
                         parents=self.trajectory.filename,
                         inline=False)
        if self.comments is not None:
            comments += self.comments

        # Results of analysis
        analysis = ""
        for x, f in self.analysis.items():
            if f is not None:
                analysis += '# %s: %s\n' % (x, f)

        # Put it all together
        # and make sure the path to the output file exists
        import os
        from atooms.core.utils import mkdir
        mkdir(os.path.dirname(self._output_file))
        with open(self._output_file, 'w') as fh:
            fh.write(comments)
            if len(analysis) > 0:
                fh.write(analysis)
            numpy.savetxt(fh, dump, fmt="%g")
            fh.flush()

    def do(self, update=False):
        """
        Do the full template pattern: compute, analyze and write the
        correlation function.
        """
        if update and not self.need_update():
            self.read()
            return

        self.compute()

        try:
            self.analyze()
        except ImportError as e:
            _log.warn('Could not analyze due to missing modules, continuing...')
            _log.warn(e.message)

        self.write()

    def __call__(self):
        self.do()
